"""
  OBSERVATION:
    
    Within this file the terms "stream" and "pipeline" have the same meaning - that of a execution engine pipeline
    While the correct term is "pipeline" as "stream" naturally refers to the data being consumed this is still used
    for backward compatibility.
"""

import os
import shutil
import json

from naeural_core import constants as ct
from copy import deepcopy
from collections import deque

from naeural_core import Logger
from naeural_core.manager import Manager
from naeural_core.config.mixins import _ConfigManagerCheckMixin
from naeural_core.config.config_manager_commands import ConfigCommandHandlers

EXTENSION = '.json'



class ConfigManager(
  Manager, 
  _ConfigManagerCheckMixin,
  ConfigCommandHandlers,
  ):

  def __init__(self, log : Logger, shmem,
               owner,
               subfolder_path=None,
               fn_app_config=None,
               folder_streams_configs=None,
               **kwargs):
    self.shmem = shmem
    self.config_app = None
    self.owner = owner
    self.dct_config_streams = None
    self.admin_pipeline_name = ct.CONST_ADMIN_PIPELINE_NAME

    self._dct_retrievers = None

    self._subfolder_path = subfolder_path
    self._fn_app_config = fn_app_config
    self._folder_streams_configs = folder_streams_configs
    
    self.__command_cache = {}

    self._path_streams = None
    super(ConfigManager, self).__init__(log=log, prefix_log='[CFGM]', **kwargs)
    return

  def startup(self):
    super().startup()
    self._dct_retrievers = self._dct_subalterns

    if self._subfolder_path is not None:
      save_path = os.path.join(self.log.get_data_folder(), self._subfolder_path)
      if not os.path.exists(save_path):
        os.makedirs(save_path)

    if self._fn_app_config is None:
      self._fn_app_config = 'config_app.txt'

    if self._folder_streams_configs is None:
      self._folder_streams_configs = 'streams'

    if self._subfolder_path is not None:
      self._fn_app_config = os.path.join(self._subfolder_path, self._fn_app_config)
      self._folder_streams_configs = os.path.join(self._subfolder_path, self._folder_streams_configs)

    self._path_streams = os.path.join(self.log.get_data_folder(), self._folder_streams_configs)
    if not os.path.exists(self._path_streams):
      os.makedirs(self._path_streams)
    return

  @property
  def cfg_app_communication(self):
    # self.config_app points to the `config_app.txt` json available in `_data/box...`
    return self.config_app.get(ct.CONFIG_APP_v2.K_COMMUNICATION, None)

  @property
  def cfg_app_file_upload(self):
    # self.config_app points to the `config_app.txt` json available in `_data/box...`
    return self.config_app.get('FILE_UPLOAD', None) ### TODO ct  
  
  
  @property
  def formatter(self):
    io_formatter_manager = self.shmem.get('io_formatter_manager', None)
    if io_formatter_manager is not None:
      formatter, formatter_name =  io_formatter_manager.get_formatter()
      return formatter
    return
  
  def _check_valid_payload_type(self, payload_type):
    assert payload_type in [
      ct.PAYLOAD_CT.COMMANDS.UPDATE_CONFIG,
      ct.PAYLOAD_CT.COMMANDS.UPDATE_PIPELINE_INSTANCE,
      ct.PAYLOAD_CT.COMMANDS.BATCH_UPDATE_PIPELINE_INSTANCE,
    ]
    if payload_type not in self.__command_cache:
      self.__command_cache[payload_type] = deque(maxlen=10)
    return
  
  
  def _add_command_to_cache(self, payload, payload_type):
    self._check_valid_payload_type(payload_type)
    self.__command_cache[payload_type].append(payload)
    return
  
  
  def _check_duplicate_last(self, payload, payload_type):
    self._check_valid_payload_type(payload_type)    
    if len(self.__command_cache[payload_type]) > 0:
      if self.__command_cache[payload_type] == payload:
        return True
    self._add_command_to_cache(payload, payload_type)
    return False
  
    

  def _get_config_plugin_instance(self, name, config):
    _module_name, _class_name, _class_def, _class_config = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_CONFIG_RETRIEVE_PLUGINS,
      name=name,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_CONFIG_RETRIEVE_PLUGINS,
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_CONFIG_PLUGINS,
    )

    if _class_def is None:
      msg = "Error loading config retrieve plugin '{}'".format(name)
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        info="No code/script defined for config retrieve plugin '{}' in {}".format(name, ct.PLUGIN_SEARCH.LOC_CONFIG_RETRIEVE_PLUGINS)
      )
    #endif

    try:
      if self.owner.runs_in_docker:
        self.P("Running in Docker - adding fallback config '{}'".format(ct.CONFIG_APP_DOCKER_FALLBACK))
        config[ct.CONFIG_APP_FALLBACK] = ct.CONFIG_APP_DOCKER_FALLBACK
      retriever = _class_def(
        log=self.log,
        config=config,
        subfolder_path=self._subfolder_path,
        fn_app_config=self._fn_app_config,
        folder_streams_configs=self._folder_streams_configs
      )
    except Exception as exc:
      msg = "Exception '{}' when initializing config retrieve plugin {}".format(exc, name)
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        autocomplete_info=True
      )
      raise exc
    #end try-except
    return retriever



  def _save_app_configuration(self, config_app):
    self.log.save_data_json(config_app, self._fn_app_config)
    return
  
  
  def _save_instance_modifications(
    self,
    pipeline_name,
    signature,
    instance_id, 
    config, 
    skip_update_time=False,
    verbose=False,
  ):
    """
    Will save the instance modifications in the local cache. 
    This is used when the instance is modified in-memory

    Parameters
    ----------
    pipeline_name : str
      The name of the pipeline
      
    signature : str
      The signature of the plugin
      
    instance_id : str
      The instance id of the plugin
      
    config : dict
      The configuration of the plugin instance    
    """
    result = False
    instance_config = self._get_plugin_instance(
      stream_name=pipeline_name,
      signature=signature,
      instance_id=instance_id,
    )
    if instance_config is not None:
      if verbose:
        self.log.P("Saving instance  <{}:{}:{}> config to local cache for following keys:\n{}".format(
          pipeline_name, signature, instance_id, json.dumps(config, indent=4)), color='b'
        )
      else:
        self.log.P("Saving instance  <{}:{}:{}> config to local cache".format(
          pipeline_name, signature, instance_id), color='b'
        )
      for k,v in config.items():
        instance_config[k] = v
      #endfor  
      self._save_stream_config(self.dct_config_streams[pipeline_name], skip_update_time=skip_update_time)
      result = True
    #endif saving config
    return result


  def _save_stream_config(self, config_stream, subfolder_path=None, prefix_fn=None, skip_update_time=False):
    if subfolder_path is None:
      subfolder_path = self._folder_streams_configs
      if not skip_update_time:
        config_stream[ct.CONFIG_STREAM.LAST_UPDATE_TIME] = self.log.now_str(nice_print=True)
    #endif
    name = config_stream['NAME']
    fname = name + EXTENSION
    if prefix_fn is not None:
      fname = prefix_fn.rstrip('_') + '_' + fname
    self.P("Saving pipeline {}:{} in {}".format(name, fname, subfolder_path))
    self.log.save_data_json(config_stream, fname, subfolder_path=subfolder_path)
    return
  


  def _delete_stream_config(self, stream_name):
    fname = stream_name + EXTENSION
    fn_full =  os.path.join(self.log.get_data_folder(), self._folder_streams_configs, fname)
    if os.path.isfile(fn_full):
      os.remove(fn_full)
      self.P(f"Pipeline '{stream_name}' deleted from local cache")
    else:
      msg = "WARNING: Seems the pipeline was already deleted: {}".format(fn_full)
      self.P(msg, color='r')      
    return


  def _save_streams_configurations(self, dct_config_streams):
    for _,config_s in dct_config_streams.items():
      self._save_stream_config(config_s)
    return


  def _flush_existing_app_configuration(self):
    path_config_app = self.log.get_data_file(self._fn_app_config)
    if path_config_app is not None:
      os.remove(path_config_app)
    return


  def _flush_existing_streams_configurations(self):
    path_streams = self.log.get_data_subfolder(self._folder_streams_configs)
    if path_streams is not None:
      # shutil.rmtree(path_streams) # destroy everything - not advisable
      lst_files_txt = [x for x in os.listdir(path_streams) if EXTENSION in x]
      for fn in lst_files_txt:
        fn_full = os.path.join(path_streams, fn)
        self.P("Deleting pipeline file '{}'".format(fn_full))
        os.remove(fn_full)
    return
  
  def __apply_nested_key(self, config: dict, nested_key: str, value: any):
    # will apply the value to the nested key in the configuration
    keys = nested_key.split('.')
    for nested_key in keys[:-1]:
      if nested_key not in config:
        config[nested_key] = {}
      config = config[nested_key]
    #endfor
    config[keys[-1]] = value
    return

  def _apply_delta_to_config(self, original_config: dict, delta_config: dict, ignore_fields: list = None) -> dict:
    """
    Will apply the delta configuration to the original configuration inplace.
    The delta configuration is a dictionary with the keys that need to be updated.
    The keys can be nested, separated by '.'. In this case, the original configuration
    will be traversed in depth.
    
    Parameters
    ----------
    original_config : dict
        The original configuration
    delta_config : dict
        The proposed changes to the original configuration
    ignore_fields : list | None, optional
        List of fields to ignore when applying the delta configuration, by default None

    Returns
    -------
    dict
        The updated configuration
    """

    if ignore_fields is None:
      ignore_fields = []

    for k, v in delta_config.items():
      if k in ignore_fields:
        continue
      
      if '.' not in k:
        original_config[k] = v
      else:
        self.__apply_nested_key(original_config, k, v)
    # endfor

    return original_config
  
  def save_instance_modifications(self, pipeline_name, signature, instance_id, config):
    """
    Will save the instance modifications in the local cache. 
    This is used when the instance is modified in-memory

    Parameters
    ----------
    pipeline_name : str
      The name of the pipeline
      
    signature : str
      The signature of the plugin
      
    instance_id : str
      The instance id of the plugin
      
    config : dict
      The configuration of the plugin instance    
    """
    return self._save_instance_modifications(pipeline_name, signature, instance_id, config)  
  
  
  def save_pipeline_modifications(self, pipeline_name, pipeline_config, skip_update_time=False):
    """
    Will save the pipeline modifications in the local cache. 
    This is used when the pipeline is modified in-memory

    Parameters
    ----------
    pipeline_name : str
      The name of the pipeline
      
    pipeline_config : dict
      The configuration of the pipeline    
    """
    if pipeline_name not in self.dct_config_streams:
      self.P("Cannot save pipeline modifications for non-existing pipeline '{}'".format(pipeline_name), color='error')
      return None
    #endif
    
    for k,v in pipeline_config.items():
      self.dct_config_streams[pipeline_name][k] = v
      
    return self._save_stream_config(self.dct_config_streams[pipeline_name], skip_update_time=skip_update_time)


  def retrieve(self, lst_config_retrieve):
    self.P("Performing config retrieve...")
    for config in lst_config_retrieve:
      plugin_name = config[ct.CONFIG_RETRIEVE.K_TYPE]
      retriever = self._get_config_plugin_instance(plugin_name, config)
      endpoint = config.get(ct.CONFIG_RETRIEVE.K_APP_CONFIG_ENDPOINT, None)
      self.P("  Running retriever {}:{}".format(plugin_name, endpoint))
      if retriever:
        self._dct_retrievers[plugin_name] = retriever
        retriever.connect(**config.get(ct.CONFIG_RETRIEVE.K_CONNECT_PARAMS, {}))

        config_app, dct_config_streams = retriever.retrieve(
          app_config_endpoint=endpoint,
          streams_configs_endpoint=config.get(ct.CONFIG_RETRIEVE.K_STREAMS_CONFIGS_ENDPOINT, None)
        )

        if config_app is not None:
          self.P("  Connected and retrieved comms config of size {}...".format(len(config_app)))
          self.P("  Running config check...")
          config_app = self._check_config_app(config_app)
          self.P("  App config - Flushing existing config and saving the new one ...")
          self._flush_existing_app_configuration()
          self._save_app_configuration(config_app)
        else:
          self.P("  No comms config retrieved from given endpoint", color='r')          
        #endif

        if dct_config_streams is not None:
          self.P("  Connected and retrieved streams config...")
          formatter = self.formatter
          if formatter is not None:
            dct_config_streams = formatter.decode_streams(dct_config_streams)
          #endif

          dct_config_streams = self.keep_all_good_streams(dct_config_streams)
          if len(dct_config_streams) > 0:
            self.P("  Streams configs - Flush existing configs and saving the new ones ...")
            self._flush_existing_streams_configurations()
            self._save_streams_configurations(dct_config_streams)
          #endif
        else:
          self.P("  No streams configured from endpoint", color='r')          
        #endif if dct-config-streams
      #endif
    #endfor
    return
  
  
  def maybe_setup_admin_pipeline(self):
    """
    
    Returns
    -------
    
    list with the names of the available pipelines
    """
    ADMIN_PIPELINE_VER = "2.1.1"
    INSTANCE_STR = "{}_INST"
    pipeline_name = self.admin_pipeline_name
    self.P("Setting-up admin jobs pipeline...")  
    # we setup the default admin pipeline type and params  
    default_admin_pipeline_setup = {
      ct.CONFIG_STREAM.K_NAME     : self.admin_pipeline_name,
      ct.CONFIG_STREAM.K_TYPE     : ct.CONFIG_STREAM.DEFAULT_ADMIN_PIPELINE_TYPE,
            
      ct.CONFIG_STREAM.PIPELINE_OPTIONS.NETWORK_LISTENER_PATH_FILTER : [
        None, None, 
        ct.ADMIN_PIPELINE_FILTER, # TODO: this should be dinaically created based on the plugins that use the DCT
        None
      ],
      ct.CONFIG_STREAM.PIPELINE_OPTIONS.NETWORK_LISTENER_MESSAGE_FILTER : {},
      ct.CONFIG_STREAM.PIPELINE_OPTIONS.ADMIN_PIPELINE_VER : ADMIN_PIPELINE_VER,
    }
    # we setup the default admin pipeline
    default_admin_pipeline = {
      **default_admin_pipeline_setup,
      ct.CONFIG_STREAM.K_PLUGINS  : [
        # no plugins in the default admin pipeline - will be added later based on the template
        ],
    }
    # get the mandatory admin pipeline template 
    # this does NOT contain instances, only the plugin signatures 
    # and their default configuration
    admin_pipeline_template = self.owner.admin_pipeline_template
    if False:
      self.P("Using admin pipeline BASE TEMPLATE configuration:\n{}".format(json.dumps(admin_pipeline_template, indent=4)))
    # this is the reference pipeline
    dct_admin_pipeline = deepcopy(default_admin_pipeline)
    # get all the mandatory plugins
    lst_admin_signatures = list(admin_pipeline_template.keys())
    # for each plugin in the template, create a new instance and add it to the "reference" admin pipeline
    for i, plg in enumerate(lst_admin_signatures):
      plgcfg = deepcopy(admin_pipeline_template[plg])
      plgcfg[ct.BIZ_PLUGIN_DATA.INSTANCE_ID] = INSTANCE_STR.format(plg)
      dct_admin_pipeline[ct.CONFIG_STREAM.K_PLUGINS].append({
        ct.BIZ_PLUGIN_DATA.SIGNATURE  : plg,
        ct.BIZ_PLUGIN_DATA.INSTANCES : [plgcfg]
      })
    #endfor all mandatory plugins injected in the "reference" admin pipeline
    
    dct_current_admin = self.dct_config_streams.get(pipeline_name)
    needs_save = False
    if dct_current_admin is None:
      # no admin pipeline found - add the reference pipeline
      # maybe this is the first time the device is started
      self.dct_config_streams[pipeline_name] = dct_admin_pipeline
      dct_current_admin = dct_admin_pipeline
      needs_save = True # save the new admin pipeline
      self.P("  No admin pipleine found. Added administration pipeline")
    else:
      admin_pipeline_dct_filter = dct_current_admin.get(ct.CONFIG_STREAM.PIPELINE_OPTIONS.NETWORK_LISTENER_PATH_FILTER, [])
      if admin_pipeline_dct_filter != default_admin_pipeline_setup[ct.CONFIG_STREAM.PIPELINE_OPTIONS.NETWORK_LISTENER_PATH_FILTER]:
        needs_save = True
      admin_pipeline_type = dct_current_admin.get(ct.CONFIG_STREAM.K_TYPE, None)
      # pipeline found - check if all mandatory plugins are present
      lst_current_signatures = [x[ct.PLUGIN_INFO.SIGNATURE] for x in dct_current_admin[ct.CONFIG_STREAM.K_PLUGINS]]
      # get all the instances of the current admin pipeline assuming single instances
      dct_curr_plugins_instances = {
        x[ct.PLUGIN_INFO.SIGNATURE] : x[ct.BIZ_PLUGIN_DATA.INSTANCES][0]  for x in dct_current_admin[ct.CONFIG_STREAM.K_PLUGINS] 
      }
      self.P("Found admin pipeline type `{}` with {}".format(admin_pipeline_type, lst_current_signatures))
      # get the "reference" admin pipeline with the default configuration in same signature: data format as the current admin pipeline
      if admin_pipeline_type != dct_admin_pipeline[ct.CONFIG_STREAM.K_TYPE]:
        self.P("Admin pipeline type is `{}` - should be `{}`".format(
          admin_pipeline_type, dct_admin_pipeline[ct.CONFIG_STREAM.K_TYPE]
        ))
        needs_save = True # update the admin pipeline type at save time
      #endif admin pipeline type is not correct
      dct_std = {x[ct.PLUGIN_INFO.SIGNATURE] : x for x in dct_admin_pipeline[ct.CONFIG_STREAM.K_PLUGINS]}
      self.P("  Required signatures from template: {}".format(lst_admin_signatures))
      for plg in lst_admin_signatures:
        # for each mandatory plugin in the template, check if it is present in the current admin pipeline
        # and if not, add it. Also check if the instance id is correct and if all the mandatory keys are present
        if plg not in lst_current_signatures:
          dct_current_admin[ct.CONFIG_STREAM.K_PLUGINS].append(dct_std[plg])
          self.P("    Missing admin plugin instance `{}` added".format(plg))
          needs_save = True  
        else:
          default_instance_config = dct_std[plg][ct.BIZ_PLUGIN_DATA.INSTANCES][0] # instance 0 is the default instance from the reference pipeline
          curr_instance_id = dct_curr_plugins_instances[plg][ct.BIZ_PLUGIN_DATA.INSTANCE_ID]
          self.P("    Checking admin plugin instance `{}`: template has {}".format(
            curr_instance_id, #list(dct_curr_plugins_instances[plg].keys()),
            list(default_instance_config.keys()), 
          ))
          correct_instance_id = INSTANCE_STR.format(plg)
          # check if the instance id is correct
          if curr_instance_id != correct_instance_id:
            dct_curr_plugins_instances[plg][ct.BIZ_PLUGIN_DATA.INSTANCE_ID] = correct_instance_id
            self.P("    Corrected instance id for {}: {} -> {}".format(plg, curr_instance_id, correct_instance_id))
            needs_save = True
          #endif instance id must respect the standard
          # now check if all the mandatory keys are present
          for k in default_instance_config:
            if k not in dct_curr_plugins_instances[plg]:
              dct_curr_plugins_instances[plg][k] = default_instance_config[k]
              self.P("      Missing {}:{} admin plugin instance key added added".format(plg, k))
              needs_save = True
            #endif missing key
          #endfor all keys              
        #endif not already admin plugin in config
      #endfor all must-in plugins
    #endif current is null or not      
    
    if needs_save:
      # we make sure we respect the default configuration
      dct_current_admin = {
        **dct_current_admin,
        **default_admin_pipeline_setup, # we add the default type, etc
      }
      self.dct_config_streams[pipeline_name] = dct_current_admin # update the admin pipeline in the memory cache      
      self.P("  Saving admin pipeline post modification:\n{}".format(json.dumps(dct_current_admin, indent=2)))
      self._save_stream_config(dct_current_admin)
      # now maybe replace the secrets
      res = self.log.replace_secrets(dct_current_admin)
      if res is not None and len(res) > 0:
        self.P("Admin pipeline secrets replaced post save: {}".format(res), color='m')
    else:
      self.P("  Admin pipeline is already correctly configured - no need to save")
    #endif needs save
    names_available_pipelines = list(self.dct_config_streams.keys())
    return names_available_pipelines
    

  def __convert_txt_to_json(self, fn_txt):
    fn_json = fn_txt.replace('.txt', EXTENSION)
    if not os.path.isfile(fn_json):
      self.P("Converting '{}' to '{}'".format(fn_txt, fn_json))
      shutil.copy(fn_txt, fn_json)
    if os.path.isfile(fn_txt):
      self.P("Removing '{}'".format(fn_txt))
      os.remove(fn_txt)
    return fn_json

  def load_streams_configurations(self):
    self.P("Loading streams from local cache...")
    dct_config_streams = {}
    available_streams = os.listdir(self._path_streams)
    available_streams = [os.path.join(self._path_streams, x) for x in available_streams]
    available_streams = list(filter(lambda x: x.endswith(EXTENSION) or x.endswith('.txt'), available_streams))
    
    for i in range(len(available_streams)):
      fn = available_streams[i]
      if fn.endswith('.txt'):
        try:
          new_fn = self.__convert_txt_to_json(fn)
          available_streams[i] = new_fn
        except:
          self.P("Cannot convert '{}' to '{}'. Skipping.".format(fn, EXTENSION), color='r')
          continue
        #endtry
      #endif .txt file        
    #endfor
    

    for fn in available_streams:
      crt_config_stream = self.log.load_json(
        fname=fn, folder=None, numeric_keys=False,
        replace_environment_secrets='$EE_',
      )
      if crt_config_stream is not None:
        crt_config_stream = self.keep_good_stream(crt_config_stream)
      else:
        self.P("Stream {} could not be loaded. Skipping.".format(fn), color='r')
      #endif

      if crt_config_stream is not None:
        stream_name = crt_config_stream[ct.CONFIG_STREAM.K_NAME]
        dct_config_streams[stream_name] = crt_config_stream

        # sometimes the name of the config file may differ from the configured name. 
        # In this case we replace the file
        good_fn = os.path.join(self._path_streams, stream_name) + EXTENSION
        if fn != good_fn:
          os.remove(fn)
          self._save_stream_config(crt_config_stream)
        #endif
      else:
        self.P("Stream {} is not correct configured. Skipping.".format(fn), color='r')
      #endif
    #endfor
    
    self.dct_config_streams = dct_config_streams

    names_available_streams = self.maybe_setup_admin_pipeline()

    if len(names_available_streams) > 0:
      self.P("Available streams: {}".format("; ".join(names_available_streams)), color='b')
    else:
      self.P("No available pipelines. At least administration pipeline should be running!", color='error')

    return
  
  def load(self):
    # loads `config_app.txt` json available in `_data/box...`
    self.P("Loading '{}' from local cache...".format(self._fn_app_config), color='b')
    self.config_app = self.log.load_data_json(
      fname=self._fn_app_config, 
      numeric_keys=False,
      replace_environment_secrets='$EE_',
      allow_missing_secrets=False,
    )
    if self.config_app is None:
      msg = "CRITICAL ERROR! The device cannot start - there is no app configuration"
      raise Exception(msg)

    # set the communication configuration in shmem to be used in a facile way in map reduce streams
    self.shmem['config_communication'] = self.cfg_app_communication

    # now lets load the streams
    self.load_streams_configurations()
    self.P("OK! The device can start", color='g')
    return 
  

  
  





if __name__ == '__main__':

  log = Logger(lib_name='A', base_folder='.', app_folder='_local_cache', TF_KERAS=False)

  lst_config = [
    {
      'TYPE' : 'remote',
      'APP_CONFIG_ENDPOINT' : 'https://www.dropbox.com/s/yjhgzd6uwuo49po/config_app_new.txt?dl=1',
      'STREAMS_CONFIGS_ENDPOINT' : 'https://www.dropbox.com/sh/9fhxp5ob9cnf82r/AAAJn5loxbhRK03t-VmsQQXMa?dl=1',
    }
  ]
  
  m = ConfigManager(
    log=log,
    subfolder_path='__test_config3'
  )
  m.retrieve(lst_config)
  m.load()



from time import sleep
import multiprocessing as mp

  
from naeural_core import Logger

if __name__ == '__main__':
  mp.set_start_method('spawn')
  
  l = Logger('PMTX', base_folder='.', app_folder='_cache')
  
  lock = l.lock_process('AID_MOB')
  
  if lock is not None:    
    print("Lock: {}".format(lock))
    for i in range(100):
      print("I'm running",i)
      sleep(2)
    print("I'm done")
  else:
    print("Process already running")

  



import pymssql
import pandas as pd

from naeural_core.local_libraries.db_conn.base import BaseConnector

class PyMSSQLConnector(BaseConnector):
  def __init__(self, **kwargs):
    self._cnxn = None
    self._default_sql_query = "SELECT * FROM {};"
    super(PyMSSQLConnector, self).__init__(**kwargs)
    return

  def _connect(self, **kwargs):
    self._cnxn = pymssql.connect(**kwargs)
    return

  def check_connection(self):
    try:
      cursor = self._cnxn.cursor()
      cursor.execute("SELECT 1")
    except pymssql.Error as e:
      return False
    return True

  def _create_reader(self, **kwargs):
    table_data = kwargs.get('table_data', '')
    chunksize = kwargs.get('chunksize', None)
    sql_query = kwargs.get('sql_query', '')
    assert len(table_data) > 0 or len(sql_query) >0, "Must either have table or SQL query"
    if len(sql_query) == 0:
      sql_query = self._default_sql_query.format(table_data)

    reader = pd.read_sql(
      sql=sql_query,
      con=self._cnxn,
      chunksize=chunksize
     )
    return reader



if __name__ == '__main__':

  from naeural_core import Logger

  log = Logger(
    lib_name='DB', base_folder='.', app_folder='_local_cache',
    TF_KERAS=False
  )
  sql_query = "SELECT * FROM [CAVI_EVENTS].[dbo].[EV_202210]"
  query_params = {'sql_query': sql_query,
      'chunksize': 1000000,}
  config = {
    'CONNECT_PARAMS' : {
      "password": "3edc$RFV",
      "server": "172.16.16.27",
      "user": "caviuser"
    },


  }

  conn = PyMSSQLConnector(log=log, config=config)
  conn.connect(nr_retries=5)
  dct_data = conn.get_data(**query_params)
  print(dct_data)


from naeural_core.business.base import SimpleRestExecutor as BasePlugin



__VER__ = '0.1.0'

_CONFIG = {
  **BasePlugin.CONFIG,
  
  'RESULT_KEY'              : 'exec_result',
  'ERROR_KEY'               : 'exec_errors',  
  'WARNING_KEY'             : 'exec_warning',
  
  'ROOT_PAYLOAD_KEY'        : 'ROOT_PAYLOAD_DATA',
  
  'DEBUG_EXEC'              : False,

  'MAX_INPUTS_QUEUE_SIZE'   : 1, # default queue set to 1 due to imposed delay 
  
  # the following two flags have been defaulted in such a way to force the plugin
  # execution only when IMG is available - this way any custom exec request will 
  # be executed only on "loaded" iterations of the plugin process
  'ALLOW_EMPTY_INPUTS'      : False, #
  'RUN_WITHOUT_IMAGE'       : False,
  
 
  "PROCESS_DELAY" : 1,
  
  "INFO_MANIFEST" : {
    "NAME" : "REST-like custom execution",

    "REQUEST" :{ 
      "DATA" : {
        "CODE" : "Base64 encoded script",
        },
      "TIMESTAMP" : "timestamp float (optional)"
      }
  },

  'VALIDATION_RULES' : {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },

}

_DEFAULT_INSTANCE = {
    "SIGNATURE": "REST_CUSTOM_EXEC_01",
    "INSTANCES": [
        {
            "INSTANCE_ID": "REST_EXEC",
        }
    ],
}


SLEEP_ON_ERRORS = 10

class RestCustomExec01Plugin(BasePlugin):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    super(RestCustomExec01Plugin, self).__init__(**kwargs)
    return
  
  

  def _on_request(self, request):
    result, errors, warnings = None, None, []
    
    if not isinstance(request, dict):
      errors = "Request does not comply with the given manifest"
    else:
      str_b64code = request.get('CODE')
      if str_b64code is not None and len(str_b64code) > 0:
        res = self.exec_code(
          str_b64code=str_b64code,
          debug=self.cfg_debug_exec,
          self_var='plugin', # we set plugin local for exec code
          modify=True,
        )
        if isinstance(res, tuple):
          result, errors, warnings = res
      else:
        errors = "Received empty CODE"

    if self.cfg_allow_empty_inputs:
      warnings.append('Result might be erroneous due to plugin being allowed to run without input')
      
    dct_root_payload_data = {}
    if isinstance(result, dict) and len(result) > 0:
      _rkey = self.cfg_root_payload_key
      rkey1 = _rkey.lower()
      rkey2 = _rkey.upper()
      dct_root = result.get(rkey1, result.get(rkey2))
      if dct_root is not None and isinstance(dct_root, dict) and len(dct_root) > 0:
        dct_root_payload_data = dct_root
      #endif root payload in result
    #endif result is dict
    kwargs = {
      self.cfg_result_key : result,
      self.cfg_error_key : errors,
      self.cfg_warning_key : warnings,
    }            
    payload = self._create_payload(
      **kwargs,
      **dct_root_payload_data,
    )    
    
    if len(warnings) > 0:
      self.P("Execution warnings: {}".format(warnings), color='r')

    if errors is not None:
      if isinstance(errors, list) and len(errors[-1]) > 100:
        trace = errors[-1]
        errors = errors[:-1]
      else:
        trace = None
      _show_err = True
      sep = '*' * 100
      msg = "Exception in {}".format(self)
      info = "Reset/sleep plugin for {}s. Error in custom code execution: {}\n{}\n{}\n{}\n{}".format(
        SLEEP_ON_ERRORS, errors, 
        sep, self.base64_to_code(str_b64code),
        sep if trace is not None else '', trace if trace is not None else ''
      )
      if _show_err:
        self.P(msg +' ' + info, color='r')
      self._create_error_notification(msg=msg, info=info, displayed=_show_err)
      self.add_payload(payload)
      payload = None # after forced send set to None so no sending again...
      self.reset_plugin_instance()
      self.sleep(SLEEP_ON_ERRORS)
    return payload
    


from naeural_core.utils.sanchez_utils.Actor import *

BABY_HEDGE_STEPS_LIMIT = 25


class Hedgehog(Actor):
  def __init__(self, name='', color=None, actor_id=1,
               x=0, y=0, alpha=0, beta=0,
               speed=0, period=0, steps=0,
               life_state=0, baby_steps=0
               ):
    super(Hedgehog, self).__init__(
      name=name, color=color, actor_id=actor_id,
      x=x, y=y, alpha=alpha, beta=beta,
      speed=speed, period=period, steps=steps,
      life_state=life_state, baby_steps=baby_steps
    )

  def get_species(self):
    return TYPE_HEDGEHOG

  def _get_baby_steps_limit(self):
    return BABY_HEDGE_STEPS_LIMIT

  def make_baby(self, env, x, y):
    baby = Hedgehog(name=self.name + 'son', color=self.color, x=x, y=y)
    env.add_actor(baby)
    self.P(f'Baby {baby.get_data_string()} was born!')

  def manage_collision(self, actor, env, x, y):
    if type(actor) == type(self):
      if actor.life_state == BABY_STATE or actor.life_state == RETIRED_STATE:
        return

      self.make_baby(env, x, y)
      self._change_direction()

    elif actor.get_species() == TYPE_SNAKE:
      if actor.life_state == BABY_STATE or actor.life_state == ADULT_STATE:
        return

      actor.lose_life(env.get_current_map())
      self._change_direction()







"""
Script:
[THTST][2021-10-06 15:15:07]  torch_script = 0.0015s, max: 0.0060s, curr: 0.0010s
[THTST][2021-10-06 15:15:39]  torch_script = 0.0015s, max: 0.0070s, curr: 0.0021s
[THTST][2021-10-06 15:15:39] 0.001549411225724626
[THTST][2021-10-06 15:15:07] 0.0015447041890523335
"""
import numpy as np
import torch as th
import torchvision as tv

from naeural_core import Logger

  

if __name__ == '__main__':
  
  l = Logger('THTST', base_folder='.', app_folder='_cache', TF_KERAS=False)
  
  train = tv.datasets.MNIST(root=l.get_data_folder(), train=True, download=True)
  test = tv.datasets.MNIST(root=l.get_data_folder(), train=False, download=True)
  
  TOP = 10
  
  
  (x_train, y_train), (x_dev, y_dev) = (train.data.numpy(), train.targets.numpy()), (test.data.numpy(), test.targets.numpy())
  x_train = (x_train.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  x_dev = (x_dev.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  y_train = y_train.astype('int64')
  y_dev = y_dev.astype('int64')
  
  l.gpu_info(True)
  
 
  model = th.jit.load('xperimental/th_basic/mnist0.sth')
  dev = next(model.parameters()).device
  th_x_dev = th.tensor(x_dev, device=dev)
  l.P("Model: \n{}".format(model))
  timings = []
  with th.no_grad():
    for _ in range(1000):
      l.start_timer('torch_script')
      th_y_pred = model(th_x_dev)
      tt = l.stop_timer('torch_script')
      timings.append(tt)
  np_yh = th_y_pred.cpu().numpy()
  
  _xd, _yp = l.load_pickle_from_data('test_mnist0')
  assert np.allclose(_xd[:TOP], x_dev[:TOP])
  assert np.allclose(_yp[:TOP], np_yh[:TOP])
  
  l.show_timers()
  np_t = np.array(timings[1:])
  l.P(np_t.mean())





import torch as th

AVAILABLE_WEIGHTS_SIZES = [4, 8, 16, 32]
DEFAULT_WEIGHTS_SIZE = 16

class _LlmUtilsMixin(object):
  """
  Mixin for LLM utilitary methods that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_LlmUtilsMixin, self).__init__()
    return

  def validate_weights_size(self, weights_size):
    """
    Validate the weights size.
    If the weights size is not in the available weights sizes and is not None, the default weights size is returned.
    If the weights size is None, it is returned as is. That means that the model will be loaded without quantization.
    Parameters
    ----------
    weights_size : int
        the weights size

    Returns
    -------
    int
        the validated weights size
    """
    if weights_size is None:
      return weights_size
    if weights_size not in AVAILABLE_WEIGHTS_SIZES:
      return DEFAULT_WEIGHTS_SIZE
    return weights_size

  def get_model_load_config(
      self, model_name, token, has_gpu=True, weights_size=None, device_map="auto",
      cache_dir=None, use_flash_attention=False
  ):
    """
    Creates the `model.from_pretrained` load config including the BitsAndBytes configuration.
    Parameters
    ----------
    model_name : str
      The model name
    token : str
      The Hugging Face token
    has_gpu : bool
      If the machine has a GPU
    weights_size : int
      The weights size
    device_map : str
      The device map
    cache_dir : str
      The cache directory
    use_flash_attention : bool
      If the Flash Attention should be used

    Returns
    -------
    tuple[dict, dict]
      the two configurations for the model and the quantization.
    """
    cache_dir = cache_dir or self.get_models_folder()
    weights_size = self.validate_weights_size(weights_size)
    torch_dtype = "auto"
    load_in_8bit = (weights_size == 8)
    load_in_4bit = (weights_size == 4)
    quantization_params = None
    self.P("Preparing model load config for {}...".format(model_name))
    self.P("  Model weights size: {}".format(weights_size))
    self.P("  Has GPU: {}".format(has_gpu))
    if has_gpu and weights_size is not None:
      if weights_size in [4, 8]:
        # We're using 4 or 8 bit quantization and we need to prepare
        # the bitsandbytes quantization arguments.
        quantization_params = {
          "load_in_8bit": load_in_8bit,
          "load_in_4bit": load_in_4bit,
        }
        if load_in_8bit:
          torch_dtype = None
          quantization_params['llm_int8_threshold'] = 6.0
        else:
          quantization_params = {
            "bnb_4bit_quant_type": "nf4",
            "bnb_4bit_use_double_quant": True,
            "bnb_4bit_compute_dtype": th.bfloat16,
            **quantization_params
          }
      else:
        # No BitsAndBytes quantization is used, thus the model is loaded in 16-bit or 32-bit precision.
        torch_dtype = th.float16
        if weights_size == 32:
          torch_dtype = th.float32
          self.P("WARNING: fp32 requested for a model that is usually fp16. This may cause OOM errors and if not will not yield better performance!", color='r')
        # endif 32 bits
      # endif 4/8 bits (quantization) or 16/32 bits (no quantization)
    # endif no config provided

    model_params = {
      'cache_dir': cache_dir,
      'token': token,
      'low_cpu_mem_usage': True,
      'torch_dtype': torch_dtype,
      'device_map': device_map
    }
    if use_flash_attention:
      model_params['attn_implementation'] = 'flash_attention_2'
    return model_params, quantization_params




"""
This plugin handles the following base functionality:
 - checks for new versions on the git server and restarts the EE if a new version is available
 - sends the config_startup.json file to the git server on request
 - receives the config_startup.json file from the git server on request
 - restarts the EE if the running time is above a certain threshold (default 2.5 days)
 
 
 
Examples:

{
  "EE_ID": "aid_hpc",
  "SB_ID": "aid_hpc",
  "ACTION": "UPDATE_PIPELINE_INSTANCE",
  "PAYLOAD": {
    "NAME": "admin_pipeline",
    "SIGNATURE": "UPDATE_MONITOR_01",
    "INSTANCE_ID": "UPDATE_MONITOR_01_INST",
    "INSTANCE_CONFIG": {
      "INSTANCE_COMMAND": {
        "COMMAND": "GET_CONFIG"
      }
    }
  },
  "INITIATOR_ID": "Explorer_xxxx",
  "EE_SIGN": "MEQCIEN1AH1NJ5A60L9xeKAa_EBliFDiBuR5-dWsqtjI6hz4AiB-zBSnyxQGRqmTud6QHGjquTDEcEQd2h4f2i1ZaKPzRQ==",   
  "EE_SENDER": "0xai_A6IrUO8pNoZrezX7UhYSjD7mAhpqt-p8wTVNHfuTzg-G",
  "EE_HASH": "47a9913c1f0b1cb5a4b29147cdcd9fd34c4235c9476c66521bbbd4ca1c84b69e"
}


{
  "EE_ID": "aid_hpc",
  "SB_ID": "aid_hpc",
  "ACTION": "UPDATE_PIPELINE_INSTANCE",
  "PAYLOAD": {
    "NAME": "admin_pipeline",
    "SIGNATURE": "UPDATE_MONITOR_01",
    "INSTANCE_ID": "UPDATE_MONITOR_01_INST",
    "INSTANCE_CONFIG": {
      "INSTANCE_COMMAND": {
        "COMMAND": "SAVE_CONFIG",
        "DATA": "eNqNVttu4zYQfa6/QlX7mPiSIrttgKJlJNrmRiJViUrWLQJCrp6elpuMqz7V32tVJQMTr7uP32WI7TeHz/27/R+Ug/VUS03Yp1WYiijNZxuh6WX8s/VsnvE7MOqAR1O3j9D+cb42E="   
      }
    }
  },
  "INITIATOR_ID": "Explorer_xxxx",
  "EE_SIGN": "MEUCIQC2B_WWthwdK2sXSCtuFJ0_YmevvDDpUaNLfDn2SF2PGgIgURac5MspmGFyJxccGOYOAVghb_gX-ogR7StG3y-9u-c=",   
  "EE_SENDER": "0xai_A6IrUO8pNoZrezX7UhYSjD7mAhpqt-p8wTVNHfuTzg-G",
  "EE_HASH": "38b557b6b673f75d7a809d637bf93420d477fdddc109a90b86c2fc7a3a57a7aa"
} 
 
"""
from naeural_core.business.base import BasePluginExecutor
import os


class UpdateCt:
  VERSION_RESTART = "VERSION UPDATE RESTART"
  FORCED_RESTART = "FORCED RESTART"

__VER__ = '0.1.0.0'

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  # default working hours should be non-stop
  'WORKING_HOURS': [], 
  # "WORKING_HOURS"       : [["08:30", "09:30"]],    
  
  "FORCE_RESTART_AFTER" : 3600 * 24 * 2,  # days restart time
  "REBOOT_ON_RESTART"   : False,
  
  "BUILD_DELAY"         : 10 * 60,

  'ALLOW_EMPTY_INPUTS'  : True,
  
  'SERVER_URL'          : '',

  "PROCESS_DELAY"       : 300,

  "VERSION_TOKEN"       : None,
  "VERSION_URL"         : None,
  
  "LOG_INFO"            : False,

  "RESTART_ON_BEHIND"   : True,

  "USE_YAML"            : False,
  "RELEASE_TAG"         : "staging",

  # debug stuff
  # end debug stuff

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}

class UpdateMonitor01Plugin(BasePluginExecutor):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(UpdateMonitor01Plugin, self).__init__(**kwargs)
    return


  def on_init(self):
    self._update_monitor_count = 0
    self.__logo = False
    self.__failed_download = False
    self.__restart_inititated = False
    self.__restart_inititated_time = None
    self.__restart_offset = None
    DEFAULT_STATE = {
      "last_update" : None,
      "last_prev"   : None,
      "last_new"    : None,
    }
    self.__state = self.cacheapi_load_json(default=DEFAULT_STATE)
    self.P("Update Monitor initialized:\nState: {}\nWorking hours: {}".format(
      self.__state, self.cfg_working_hours
    ))

    self.__maybe_config_supervisor_restart_offset()

    return

  def __maybe_config_supervisor_restart_offset(self):
    """Configures the restart offset for the supervisor node"""
    if not self.is_supervisor_node:
      return
    self.__restart_offset = 36 * 3600 + self.np.random.randint(0, 36 * 3600)  # 36 hours + random
    restart_time = self.time() + self.__restart_offset
    restart_time_normalized = restart_time % (24 * 3600)
    offset = 30 * 60  # 30 minutes offset
    upper_bound_normalized = (self.netmon.epoch_manager.get_current_epoch_end() + offset) % (24 * 3600)
    lower_bound_normalized = (self.netmon.epoch_manager.get_current_epoch_end() - offset) % (24 * 3600)

    if restart_time_normalized < upper_bound_normalized and restart_time_normalized > lower_bound_normalized:
      self.__restart_offset += 3 * 3600  # add 3 hours to the offset
      self.P("Supervisor node restart offset configured to {} seconds".format(self.__restart_offset))
    # endif restart time is within the epoch bounds
    return

  def __get_config_startup_as_base64(
    self, 
    compress=True, 
  ):
    """Loads the config_startup.json file and generates a base64 encoded string"""
    config_file = self.log.config_file
    if not os.path.isfile(config_file):
      raise ValueError("Config file '{}' not found!".format(config_file))
    self.P("Loading and sending config file {}".format(config_file))
    with open(config_file, 'r') as f:
      str_config = f.read()
    str_b64 = self.string_to_base64(str_config, compress=compress)
    return str_b64
  
  def _send_config_payload(self):
    # TODO: upgrade this method to send the config payload to the target_id with simmetric encrypton
    #       and decryption on the other side (the initiator or "target_id" side)
    str_config_startup_b64 = self.__get_config_startup_as_base64()
    result = self.add_payload_by_fields(
      config_startup=str_config_startup_b64,      
    )
    return
  
  
  def __validate_config(self, dct_config):
    """Validates the config_startup.json file"""
    is_ok = True
    mandatory_keys = self.ct.CONFIG_STARTUP_MANDATORY_KEYS
    for k in mandatory_keys:
      if (k not in dct_config) or (dct_config[k] is None):
        is_ok = False
      elif isinstance(dct_config[k], (list, str, dict)) and len(dct_config[k]) == 0:
        is_ok = False
    # endfor mandatory keys
    return is_ok
  
  
  def __save_config_startup(
    self, 
    str_input : str, 
    decompress=True,
  ):
    """Saves the config_startup.json file from a base64 encoded string or straight json"""
    config_file = self.log.config_file
    self.P("Saving received config in {}".format(config_file))

    is_ok = True
    fail_reason = None

    if str_input.startswith('{'):
      str_json = str_input
    else:
      str_json = self.base64_to_str(str_input, decompress=decompress)
    try:
      dct_config = self.json_loads(str_json)
      is_ok = True
    except Exception as e:
      is_ok = False
      fail_reason = str(e)

    # now validate that the received config contains the required fields
    if is_ok:
      is_ok = self.__validate_config(dct_config)
      if not is_ok:
        fail_reason = "Validation failed! Missing mandatory keys or wrong value types."
    # endif validation

    if is_ok:
      with open(config_file, 'w') as f:
        f.write(str_json)
      self.P("Node config saved OK!", boxed=True, box_char='*')
    else:
      self.P("Node config save failed: {}".format(fail_reason), color='r')
    return
  
  def _process_config_payload(self, data):
    # TODO: upgrade this method to decrypt the config payload received from the other side
    #       and update the local config with the received values
    if isinstance(data, dict):
      str_data = self.safe_json_dumps(data)
    elif isinstance(data, str):
      str_data = data
    else:
      raise ValueError("Invalid data type: {}".format(type(data)))
    self.__save_config_startup(str_input=str_data)
    return
  
  def _send_whitelist_payload(self):
    lst_allowed = self.bc.get_whitelist()
    result = self.add_payload_by_fields(
      ee_whitelist=lst_allowed,      
    )
    return
  
  
  def _send_pipelines_payload(self):
    my_pipelines = self.node_pipelines
    self.P(f"Sending {len(my_pipelines)} pipelines req by '{self.modified_by_id}' <{self.modified_by_addr}>...")
    result = self.add_payload_by_fields(
      ee_pipelines=self.node_pipelines,
      ee_is_encrypted=True,      
    )
    return
  
  
  ###

  def on_command(self, data, **kwargs):
    self.P("UPDATE_MONITOR on {} received:\nFROM: '{}' <{}>\nINIT:  '{}' <{}>\nDATA:  '{}'".format(
      self.modified_by_id, self.modified_by_addr,
      self.initiator_id, self.initiator_addr,
      self.eeid, 
      str(data)[:100]
    ))
    command = 'UPDATE_CHECK'
    target_id = None
    if isinstance(data, dict):
      data = {k.upper():v for k,v in data.items()}
      # target_id = data.get('TARGET_ID') # Obsolete
      command = data.get('COMMAND', "UPDATE_CHECK").upper()

    if command == 'UPDATE_CHECK': 
      self.P("Running on-demand git ver check and validation {} ...".format(
        self._update_monitor_count)
      )
      self._process_update_check()
    #end update_check command
    
    elif command == 'GET_CONFIG':
      self.P("Running on-demand config send...")
      self._send_config_payload()
    #end get_config command
    
    elif command == 'SAVE_CONFIG':
      self.P("Running on-demand config saving...")
      data = data.get('DATA') 
      self._process_config_payload(data=data)
    #end save_config command      
    
    elif command == 'GET_WHITELIST':
      self.P("Running on-demand whitelist request ...")
      self._send_whitelist_payload()
      
    elif command == 'GET_PIPELINES':
      self.P("Running on-demand pipelines request ...")
      self._send_pipelines_payload()
    #endif commands
    return


  def needs_forced_restart(self):
    if self.__restart_offset and self.get_node_running_time() > self.__restart_offset:
      return True

    if self.cfg_force_restart_after is None:
      return False
    elif self.get_node_running_time() > self.cfg_force_restart_after:
      if self.get_node_running_time() <= 24 * 3600:
        self.P(
          f"Forced restart required after {self.cfg_force_restart_after}s on node already running for {self.get_node_running_time()}s. Please check your configuration!", 
          color='r'
        )
      return True
    return False
  
    
  

  def get_int_ver(self, ver):
    try:
      digits = [int(x) for x in ver.split('.')]
      int_ver = 0
      for i, digit in enumerate(reversed(digits)):
        int_ver += (1000 ** i) * digit
    except:
      str_error = self.trace_info()
      msg = "Error decoding version string: {}\n{}".format(ver, str_error)
      self.P(msg, color='r')
      raise ValueError("`get_int_ver FAILED: " + msg)
    return int_ver

  def get_version_from_raw(self, resp):
    """
    Assume the response is a string with the version in the format:
    version="__VER__"

    Parameters
    ----------
    resp : str
        The response string

    Returns
    -------
    str
        The version string.
    """
    ver = resp.split('=')[1]
    ver = ver.rstrip().lstrip().replace('"','').replace("'", '')
    return ver

  def get_version_from_yaml(self, resp):
    """
    Parse the response as a yaml file and extract the version from the release tag.
    The response is expected to be in the format:
    application:
      envs:
        - name: staging
          version: "__VER__"
          info: "..."
        - name: production
          version: "__VER__"
          info: "..."
        ...

    Parameters
    ----------
    resp : str
        The response string

    Returns
    -------
    str
        The version string.
    """
    dct = self.yaml.safe_load(resp)
    lst_releases = dct["application"]["envs"]
    release = [x for x in lst_releases if x["name"] == self.cfg_release_tag][0]
    ver = release["version"]
    ver = ver.rstrip().lstrip().replace('"','').replace("'", '')

    return ver

  def get_git_server_version(self):
    ver = None
    branch = self.docker_branch if self.runs_in_docker else self.log.git_branch
    token = self.cfg_version_token
    url0 = self.cfg_version_url
    if isinstance(url0, str) and url0.startswith('http'):
      if '{}' in url0:
        url0 = url0.format(branch)
      resp = None
      try:
        headers = {}
        if token not in [None, "", "git_version_access_token", "token_for_accessing_private_repositories"]:
          headers = {'Authorization': 'token ' + token}      
        self.P("Retrieving version with url: {} and headers: {}".format(url0, headers))
        resp0 = self.requests.get(url0, headers=headers)
        status_code = resp0.status_code
        resp = resp0.content.decode()
        self.P("Retrieved for `{}:{}` status: {} data:\n{}".format(
          "docker" if self.runs_in_docker else "git-src", branch,
          status_code,
          resp[:100].replace('\n', ' '))
        )
        if resp0.status_code == 200:
          if self.cfg_use_yaml:
            ver = self.get_version_from_yaml(resp)
          else:
            ver = self.get_version_from_raw(resp)
        else:
          self.P("Failed to retrieve version with url:{}, headers:{}. Response status: {} was: {}".format(
            url0, headers, status_code, resp), color='r'
          )
      except Exception as exc:
        self.P("Exception while retrieving version with url:{} , headers:{}. Response was: {}, exception: {}".format(
          url0, headers, resp, exc), color='r')
    else:
      self.P(f"Skipping version check: {url0=}, {token=}", color='r')
    return ver  
  
  
  def _maybe_raise_failed_update_alert_and_save(self, prev_ver, new_ver):
    stored_prev_ver = self.__state.get('last_prev')
    stored_new_ver = self.__state.get('last_new')
    stored_update_time = self.__state.get('last_update')
    send_notif = False
    str_time = self.time_to_str()
    if prev_ver == stored_prev_ver and new_ver == stored_new_ver:
      msg = f"{self._signature} detected a previous FAILED update attempt on {stored_update_time}. "
      send_notif = True
    else:
      msg = f"Saving update initiation for prev:{prev_ver} new:{new_ver} at {str_time}"
    self.P(msg, color='r' if send_notif else None)
    self.__state['last_prev'] = prev_ver
    self.__state['last_new'] = new_ver
    self.__state['last_update'] = str_time
    self.cacheapi_save_json(self.__state)
    if send_notif:
      self._create_abnormal_notification(
        msg=msg,
      )
      self.add_payload_by_fields(
        status=msg,
        is_alert=True,
        is_new_raise=True,
      )
    return


  def _process_update_check(self):
    needs_restart = False
    self._update_monitor_count += 1
    
    branch = self.docker_branch if self.runs_in_docker else self.log.git_branch
    token = self.cfg_version_token
    url0 = self.cfg_version_url
        
    self.P("Running git ver check{} at {} runnig time. Forced restart: {}. YAML: {}. URL/Branch/Token: {} / {} / {}".format(
      self._update_monitor_count, self.get_node_running_time_str(), self.needs_forced_restart(), self.cfg_use_yaml,
      url0, branch, token
    ))
    server_ver = self.get_git_server_version()
    self.P("  Received: {}".format(server_ver))
    
    if not self.__logo:
      msg = "============================================================\n"
      msg+= "Running git ver check and validation {} received: {}\n".format(
        self._update_monitor_count, server_ver)
      msg+= "============================================================"   
      self.__logo = True
      self.P("Initial version check\n{}".format(msg))
    #endif log info 
    
    local_ver = self.ee_ver
    if server_ver is None:
      self.__failed_download = True
    else:
      self.__failed_download = False
      i_sver = self.get_int_ver(server_ver)
      i_lver = self.get_int_ver(local_ver)      
      if local_ver != server_ver:
        msg = "Current version {}({}) differs from server '{}' ver {}({}).".format(
          local_ver, i_lver,
          self.log.git_branch,
          server_ver, i_sver,
        )
              
        if self.__restart_inititated:
          self.P("WARNING: E2 restart already initiated at {}".format(self.__restart_inititated_time), color='r')
        elif i_sver > i_lver:
          full_msg = "Local version is below server version. E2 update is required! {}".format(msg)
          if self.cfg_restart_on_behind:
            full_msg = "Restarting E2 initiated for auto-update: " + full_msg
          self.P(full_msg, color='r')
          self.add_payload_by_fields(
            local_ver=local_ver,
            server_ver=server_ver,
            message=full_msg,
            status=msg + " INITIATED",
            is_alert=True,
            is_new_raise=True,
            restart_initiated=self.cfg_restart_on_behind,
            restart_now=False,
            forced_restart=False,
            forced_restart_running_time=self.get_node_running_time_str(),            
          )
          # now save the update initiation and check for potential previous failed update
          self._maybe_raise_failed_update_alert_and_save(prev_ver=local_ver, new_ver=server_ver)
          
          if self.cfg_restart_on_behind:
            needs_restart = True            
          #endif restart on behind            
        else:
          self.P("{} Local version is synced or better than server. All good.".format(msg))
        #endif server version is higher or lower
      #endif versions differs
    #endif server version is None

    
    if needs_restart or self.needs_forced_restart():
      msg = UpdateCt.VERSION_RESTART if needs_restart else UpdateCt.FORCED_RESTART
      self.P(f"INITIATING NODE REBOOT DUE TO <{msg}>...", color='r', boxed=True, box_char='*')
      skip = False
      
      if not self.needs_forced_restart() and (self.cfg_working_hours is None or len(self.cfg_working_hours) == 0):
        # now we need to wait as this instance is configured to restart at first sign!
        # we dont wait if the instance is configured to restart at certain uptime and is outside working hours
        t_start = self.time()
        build_delay = int(self.cfg_build_delay // 2 + self.np.random.randint(1, self.cfg_build_delay // 2))
        sleep_time = max(build_delay // 30, 10)
        while (self.time() - t_start) <= build_delay:                
          shown = False
          if skip:
            break
          t_sleep = self.time()
          while (self.time() - t_sleep) <= sleep_time:
            if self.done_loop:
              self.P("EXTERNAL STOP RECEIVED. EXITING...", color='r', boxed=True, box_char='*')
              skip = True
              break
            elapsed = self.time() - t_start
            remaining = build_delay - elapsed                  
            self.sleep(0.1)
            if not shown:
              self.P("Delayed RESTART in {:.1f}s ...".format(remaining), color='r', boxed=True, box_char='*')                
              shown = True
            #display first iter
          #endwhile delay in delay :)
        #endwhile delay
      #endif working hours if around the clock
      
      if not skip or self.needs_forced_restart():
        use_reboot = self.cfg_reboot_on_restart
        if use_reboot:
          self.cmdapi_restart_current_box()
        else:
          self.cmdapi_stop_current_box()
        #endif use restart
        self.__restart_inititated = True
        self.__restart_inititated_time = self.time_to_str()
        status = "E2 Node reboot initiated at {} based on `{}`".format(
          self.__restart_inititated_time,
          msg,
        )
        self.add_payload_by_fields(
          local_ver=local_ver,
          server_ver=server_ver,
          message=status,
          status=msg,
          is_alert=True,
          is_new_raise=True,
          restart_initiated=self.cfg_restart_on_behind,
          restart_now=True,
          forced_restart=self.needs_forced_restart(),
          forced_restart_running_time=self.get_node_running_time_str(),                      
        )
        self.P(status, color='r', boxed=True, box_char='*')
      #endif skip restart due to external stop
    #endif needs restart

    if self.__failed_download:
      self.P("WARNING: version check failed!", color='r')    
    
    return


  def process(self):
    self._process_update_check()
    return



import numpy as np

from naeural_core.utils.sensors.base import AbstractSensor
from naeural_core.local_libraries.ts.artificial_data import generate_artificial_series


_SEED_HARD = 1785

_SEED_SIMPLE =  4318


class VirtualTimeSeries(AbstractSensor):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  
  def _setup_connection(self, 
                        n_series=1,
                        n_steps=400, 
                        freq='D',
                        test_size=0.1,
                        max_value=30,
                        random_peaks=False,
                        encode_date_time=_SEED_SIMPLE,
                        no_random=None,
                        noise_size=10,
                        **kwargs,
                        ):
    if no_random is not None:
      self._random_seed = no_random
      np.random.seed(self._random_seed)
      
    else:
      self._random_seed= np.random.randint(1, 9999)
      np.random.seed(self._random_seed)
      
    dct_data = generate_artificial_series(
      n_series=n_series, 
      n_steps=n_steps, 
      freq=freq, 
      max_sales=max_value,
      noise_size=noise_size,
      random_peaks=random_peaks,
      )
    
    np_series = dct_data['SIGNAL']
    n_test = int(np_series.shape[-1] * test_size)    
    self._covar = dct_data['COVAR']
    self._periods = dct_data['PERIODS']
    self._data_train = np_series[:,:-n_test]
    self._data_test = np_series[:,-n_test:]
    self._pos = 0
    return
  
  
  def _get_single_observation(self):
    data = self._data_test[:, self._pos]
    self._pos += 1
    if self._pos >= self._data_test.shape[1]:
      self._pos = 0
    return data
  
    
  def _has_training_data(self):
    return True
    
  
  def _get_training_data(self):
    return self._data_train
  
  

if __name__ == '__main__':
  
  import matplotlib.pyplot as plt

  data_type = 'D'
  n_steps = 400 if data_type == 'D' else 1500
  n_series = 1 
  max_value = 30
  sensor = VirtualTimeSeries( 
    no_random=608, #   # Tests: 3306, 608, 5679, 7981
    n_steps=n_steps,
    n_series=n_series,
    freq=data_type,
    random_peaks=True,
    max_value=max_value,
    test_size=0.5,
    ) 
  
  
  
  train = sensor.get_train_data()
  xtrn = range(train.shape[-1])
  alert = train.max() * 1.2

  test = sensor._data_test
  xtst = range(train.shape[-1], train.shape[-1] + test.shape[-1])
  plt.figure(figsize=(26,8))
  plt.plot(xtrn, train.ravel(), 'D-', c='blue', markersize=12)
  plt.plot(xtst, test.ravel(), 'D-', c='green', markersize=12)
  # plt.bar(xtrn, train.ravel(), color='blue')
  # plt.bar(xtst, test.ravel(), color='green')
  plt.hlines(y=alert, xmin=train.shape[-1], xmax=n_steps, color='red')
  plt.title(str(sensor._random_seed))  
  plt.show()
  print(sensor._random_seed)
  
    
  


def hamming_distance(x, y):
  """Calculate the hamming distance (number of bits different) between the
  two integers given.
  >>> [hamming_distance(x, 15) for x in [0, 8, 10, 12, 14, 15]]
  [4, 3, 2, 2, 1, 0]
  """
  return bin(x ^ y).count('1')


def levenshtein_distance(s1, s2):
  if len(s1) < len(s2):
    return levenshtein_distance(s2, s1)

  # len(s1) >= len(s2)
  if len(s2) == 0:
    return len(s1)

  previous_row = range(len(s2) + 1)
  for i, c1 in enumerate(s1):
    current_row = [i + 1]
    for j, c2 in enumerate(s2):
      insertions = previous_row[j + 1] + 1  # j+1 instead of j since previous_row and current_row are one character longer
      deletions = current_row[j] + 1  # than s2
      substitutions = previous_row[j] + (c1 != c2)
      current_row.append(min(insertions, deletions, substitutions))
    previous_row = current_row

  return previous_row[-1]


def levenshtein_distance_normalised(s1, s2):
  return levenshtein_distance(s1, s2) / max(len(s1), len(s2))



import traceback
import numpy as np

from time import time, sleep
from collections import deque

from naeural_core import constants as ct

class _BasePluginLoopMixin(object):
  def __init__(self):
    super(_BasePluginLoopMixin, self).__init__()

    self.has_exec_error = False
    self.has_loop_error = False
    self.first_error_time = None
    self.first_error_info = None
    self.last_error_time = None
    self.last_error_info = None


    self.__loop_paused = False
    self.__plugin_loop_errors = 0
    self.__plugin_exec_errors = 0
    self.__post_error_ok_count = 0    
    self.__crash_simulated = False
    self.__last_error_inputs = None

    self.__lap_times = deque(maxlen=100) # process to process time - short queue to see dynamic in time
    self.__plugin_resolution = None
    

    self.__exec_times = deque(maxlen=100) # exec time in plugin_loop - short queue to see dynamic in time
    self.__exec_loop_check_time = time() # reset exec loop check time
    self.__last_execution_chain_ts = None
        
    self.__exec_counter = 0
    self.__current_iteration = 0
    self.__exec_counter_after_config = 0
    self.is_outside_working_hours = True
    
    return


  
  ############################################################
  # Public properties & methods
  ############################################################
  @property
  def actual_plugin_resolution(self):
    return self.__plugin_resolution
  
  
  @property
  def current_process_iteration(self):
    """
    Returns the current process iteration
    """
    return self.__current_iteration

  @property
  def current_exec_iteration(self):
    """
    Returns the current loop exec iteration
    """
    return self.__exec_counter

  @property
  def loop_timings(self):
    return [round(x, 2) for x in list(self.__lap_times)]

  @property
  def exec_timestamp(self):
    return self.__last_execution_chain_ts
  
  @property
  def n_plugin_exceptions(self):
    return self.__plugin_loop_errors

  @property
  def loop_paused(self):
    return self.__loop_paused
  
  @loop_paused.setter
  def loop_paused(self, val: bool):
    self.__loop_paused = val  
    return
        
  def pause_loop(self):
    self.loop_paused = True
    return
  
  def resume_loop(self):  
    self.loop_paused = False
    return
  
  def reset_exec_counter_after_config(self):
    self.__exec_counter_after_config = 0
    return

  ############################################################
  # End public properties & methods
  ###########################################################    
  
  
  def _recalc_plugin_resolution(self):
    if self.last_process_time is not None:
      this_lap = self.time_from_last_process
      self.__lap_times.append(this_lap)
      self.__plugin_resolution = round(1 / np.mean(self.__lap_times), 1)
    return  
  

  def _maybe_simulate_exec_crash(self):
    if isinstance(self.cfg_simulate_crash_exec_iter, int) and self.cfg_simulate_crash_exec_iter > 0:
      if self.__exec_counter == 1:
        self.P("Crash simulation active at iteration {}".format(
          self.cfg_simulate_crash_exec_iter),
          color='r'
        )
      if self.__exec_counter > self.cfg_simulate_crash_exec_iter and not (self.cfg_simulate_crash_exec_single and self.__crash_simulated):
        self.__crash_simulated = True
        self.P("Simulating crash in code execution...")
        raise ValueError("Test Exception")
    return
  
  
  def _maybe_simulate_loop_crash(self):
    if self.cfg_simulate_crash_loop:
      self.P("Simulating crash in plugin loop...")
      raise ValueError("Test Loop Exception")
    return
    

  def check_loop_exec_time(self):
    # check each 30s and if enough data is gathered
    if (time() - self.__exec_loop_check_time) > 30 and len(self.__exec_times) > 5:
      self.__exec_loop_check_time = time()
      # TODO: finish checking loop times!
      times = self.__exec_times
    return


  def __maybe_handle_exec_errors(self, sleep_time):
    # `has_exec_error` is triggered on exec exception
    # `has_loop_error` is triggered on loop exception (maybe a exception within on_command)
    if self.has_exec_error or self.has_loop_error:    
      nr_errors = self.__plugin_exec_errors if self.has_exec_error else self.__plugin_loop_errors
      sleep_time = ct.PLUGIN_EXCEPTION_DELAY * nr_errors
      # now we decrease each time we hit this point and no exec error occured in loop (except is below this point)
      msg = "Forcing plugin loop delay to {}s due to {} exception nr. {}. Lowering in {} good iteration(s).".format(
        sleep_time, 
        'EXEC' if self.has_exec_error else 'LOOP',
        self.__plugin_exec_errors if self.has_exec_error else self.__plugin_loop_errors, 
        self.__post_error_ok_count,
      )
      info = None
      self.__post_error_ok_count -= 1 

      if self.__post_error_ok_count <= 0:
        self.has_exec_error, self.has_loop_error = False, False
        msg_back = " Further delay cleared due to normal behaviour. Resuming normal plugin functioning."
        self.__post_error_ok_count = 0
        msg = msg + msg_back
      elif self.last_error_time is not None and self.last_error_info is not None:
        info = "Last exec error: {}:{}".format(
          self.last_error_time, "\n".join(self.last_error_info.split('\n')[-4:]),
        ) 
      #endif stop error

      self.P(msg, color='r')
      
      self._create_error_notification(
        msg=msg, 
        notif_code=ct.NOTIFICATION_CODES.PLUGIN_DELAYED,
        displayed=True,               
        info=info,
      )        
    #endif either exec or loop error   
    return sleep_time   

  def __mark_exec_error_time_and_info(self):
    self.__plugin_exec_errors += 1
    str_time = self.log.now_str(nice_print=True)
    info = traceback.format_exc()
    if self.first_error_time is None:
      self.first_error_time = str_time
      self.first_error_info = info
    self.last_error_time = str_time
    self.last_error_info = info
    return


  def _on_data(self):
    # this can be overwritten in child classes but will not benefit from various
    # mechanisms available in the pre-defined process
    self.high_level_execution_chain()
    return

  def _on_idle(self):
    # this can be overwritten in child classes but will not benefit from various
    # mechanisms available in the pre-defined process
    bool_received_input = self.dataapi_received_input()
    bool_run = bool_received_input or (not bool_received_input and self.cfg_allow_empty_inputs)
    if not bool_run:
      return
    self.high_level_execution_chain()
    return  
  
  
  

  def execute(self):
    """
    The main execution of a plugin instance.
    This public method should NOT be overwritten under any circumstance

    Returns
    -------
    None.

    """
    try:
      # testing
      self._maybe_simulate_exec_crash()
      # end testing

      # extra & "AUTO_..." processing
      self._maybe_exec_auto_processes()
      # end extra and "AUTO_..." processing

      if self.input_queue_size > 0:
        self.inputs = self.get_next_avail_input() # setting _inputs
        self._on_data()
        self._last_data_input = time()
      else:
        self._maybe_alert_on_no_data()
        self._on_idle()
      #endif data or idle
      self.config_changed_since_last_process = False

      # `__last_execution_chain_ts` moved here from `high_level_execution_chain`
      # as `high_level_execution_chain` may be missing/overwritten in some
      # inherited implementations
      self.__last_execution_chain_ts = self.log.now_str(nice_print=True, short=False)

      # payload queue addition - this is valid only for plugins that do not directly
      # use `add_payload` - do NOT use both `return payload` as well as `add_payload`
      # as below code will duplicate the payload (or use if so designed)
      payload = self.get_payload_after_exec()
      self.add_payload(payload)
      # end payload addition

      # if above code results no payload then maybe resend status
      self._maybe_resend_last_payload()
      # end resend status

      self._cmdapi_send_commands()

      # we also log this exec time
      self.last_exec_time = time()

    except:
      msg = "CRITICAL ERROR in business plugin EXECUTE: {}, Queue: {}, {}".format(
        self.log.get_error_info(), self.input_queue_size,
        "Inputs not None" if self.inputs is not None else "No inputs"
      )
      # now we record the exception inputs
      self.__last_error_inputs = self.inputs
      info = traceback.format_exc()
      self.P(msg + '\n' + info, color='r')
      self._create_error_notification(
        msg=msg,
        info=info,
        displayed=True,
      )
      self.has_exec_error = True
      self.__post_error_ok_count = 5
      self.__mark_exec_error_time_and_info()

      if not self._threaded_execution_chain:
        delay = ct.PLUGIN_EXCEPTION_DELAY
        msg = "Forcing NON-threaded business plugin process delay extension to {}s due to exception".format(delay)
        self.P(msg, color='r')
        self._instance_config[ct.PROCESS_DELAY] = delay
        self._create_error_notification(
          msg=msg,
          displayed=True,
        )
    # finally we "consume" the inputs
    self.inputs = None
    return

  def high_level_execution_chain(self):
    """
    Standard processing cycle
    """
    self._cmdapi_refresh()
    self._payload = None

    self.__current_iteration += 1

    # now preproc, proc and postproc iteration
    self.pre_process_wrapper()

    # Initializing the params for DatasetBuilderMixin
    self._dataset_builder_info = None

    self._init_dataset_builder_params()
    # Here we gather data for DatasetBuilderMixin from the current iteration's inferences
    # This is done here instead of on_data because the inferences are retrieved in
    # the pre_process_wrapper method
    self._maybe_ds_builder_gather_before_process()
    self._maybe_ds_builder_save()

    self.process_wrapper()
    self.post_process_wrapper()
    # end actual processing

    # now we can check if we need send the payload
    payload = self.get_payload_after_exec()
    if payload is not None:
      dct_payload = self.add_payload(payload)
      # testing and benchmarking
      if self._testing_manager is not None:
        # TODO: AID & Bleo -- code review end-to-end
        self._maybe_register_payload_for_tester(payload=dct_payload)
        dct_payload = self._maybe_generate_testing_results_payload(payload=dct_payload)
      # end testing and benchmarking
    #endif

    ### next lines are backwards compat with SaveImageHeavyOps
    if self.cfg_debug_save:
      self._prepare_debug_save_payload()

    return  
  

  def __on_command(self, data, **kwargs):
    if self.cfg_instance_command_log:
      self.P("* * * * * Received INSTANCE_COMMAND: <{} ...> with kwargs {}".format(str(data)[:30], kwargs))
    self._maybe_simulate_loop_crash()
    use_local_comms_only=kwargs.get('use_local_comms_only', False)
    try:
      self._on_command(data, **kwargs)
      notif = ct.STATUS_TYPE.STATUS_NORMAL
      notif_code = ct.NOTIFICATION_CODES.PLUGIN_INSTANCE_COMMAND_OK
      msg = "Command {} executed.".format(data)
      info = ""
    except:
      notif = ct.STATUS_TYPE.STATUS_EXCEPTION
      notif_code = ct.NOTIFICATION_CODES.PLUGIN_INSTANCE_COMMAND_FAILED
      info = traceback.format_exc()
      msg = "Command {} failed.".format(data)

    self._create_notification(
      notif=notif,
      notif_code=notif_code,
      msg=msg,
      info=info,
      displayed=True,
      use_local_comms_only=use_local_comms_only,
    )

    return

  def __maybe_trigger_instance_command(self):
    triggered = False
    if self.cfg_instance_command is not None and len(self.cfg_instance_command)>0:
      # new command!
      command = self.cfg_instance_command
      # make sure we save config before code
      self.archive_config_keys(keys=['INSTANCE_COMMAND'], defaults=[{}])

      # get from instance command the standard COMMAND_PARAMS object
      command_kwargs = {}
      if isinstance(command, dict):
        command_kwargs = command.pop('COMMAND_PARAMS', {})
      # make the kwargs case insensitive
      processed_kwargs = {k.lower():v for k,v in command_kwargs.items()}
      # end get command params
      
      self.__on_command(command, **processed_kwargs)
      # copy, erase and save
      triggered = True
    #end has instance command
    return triggered  
  
  def __on_init(self):
    self.P("Running build-in & custom on_init events...")
    if self.cfg_disabled:
      self.P(f"WARNING: This plugin instance of `{self.__class__.__name__}` is DISABLED", boxed=True, color='r')
      return
    if self.cfg_dataset_builder:
      self.P(f"WARNING: Dataset builder active for `{self.__class__.__name__}`", boxed=True, color='r')      
    try:
      self._on_init()
    except Exception as e:
      self.P(f"START FAILURE: {e}", boxed=True, color='r')
      self._create_error_notification(
        msg=f"Custom on_init event failed: {e}",
        displayed=True,
      )
      raise e
    self._init_process_finalized = True
    return
  

  def plugin_loop(self):
    """
    This is BusinessPlugin main execution loop (plugin loop)

      - plugin.outside_working_hours and plugin.is_process_postponed need to be handled also
      - main thread execution actually is wrapped in the "execute"
      
      stop precedence:
      PROCESS_DELAY > FORCE_PAUSE > WORKING_HOURS
      
    """
    try: 
      plugin_loop_initialized = False
      self.__on_init()
      plugin_loop_initialized = True
    except Exception as exc:
      info = traceback.format_exc()
      self.P(f"CRITICAL ERROR in plugin loop initialization: {exc}. Resuming to plugin cleanup:\n{info}", color='r')
    
    if plugin_loop_initialized:
      self.P("Thread initialized.", color='g')
      while not self.done_loop:
        try:
          # START postpone area
          if self.__loop_paused: # triggered when updating config
            sleep(0.001)          
            continue
          # INSTANCE_COMMAND: if no imposed parallel updating
          # then we can check for instance commands and trigger callback
          self.__maybe_trigger_instance_command()
          # END INSTANCE_COMMAND
          
          # PROCESS_DELAY mechanism
          if self.is_process_postponed:  
            # skip if postpone is required
            sleep(0.001)
            continue
          # END PROCESS_DELAY mechanism
          
          # FORCE_PAUSE mechanism
          if self.is_plugin_temporary_stopped: 
            sleep(0.001)
            continue
          # END FORCE_PAUSE mechanism

          self.is_outside_working_hours = self.outside_working_hours
          if self.is_outside_working_hours: # WORKING_HOURS mechanism
            # always skip if outside hours
            sleep(0.001)
            continue
          # END postpone area
          # no postponing, now outside of scheduling and no pausing so we proceed
          # to execution and save last processing time (written in process wrapper)
          _last_run = self.last_process_time # cycle timing
          start_it = time()
          self._plugin_loop_in_exec = True
          self.__exec_counter += 1
          self.execute() # further inside process_wrapper  __last_process_time is resetted
          self.__exec_counter_after_config += 1
          # send "all-good with instance" after 1st post-config iter
          if self.__exec_counter_after_config == 1:
            msg = " >>>>>>> Instance {} exec ok post config (itr {}). <<<<<<<".format(
              self, self.__exec_counter
            )
            self.P(msg)
            self._create_notification(
              notif_code=ct.NOTIFICATION_CODES.PLUGIN_CONFIG_OK,
              msg=msg,              
              plugin_running=True,
              displayed=False
            )
          #endif send "all-good with instance" after 1st post-config iter 
          self._plugin_loop_in_exec = False
          if self.last_process_time == _last_run:
            # something was wrong in execution and no one has resetted __last_process_time
            # so we can postpone/delay if needed (otherwise process delay will not work)
            self._reset_last_process_time()
          #endif
          it_time = time() - start_it
          # add this exec iteration time to the queue
          self.__exec_times.append(it_time)
          # now we can check if something strange is happening in the loop such as a
          # abnormal duration of the execute
          self.check_loop_exec_time()
          if self.cfg_forced_loop_sleep is not None:
            sleep_time = self.cfg_forced_loop_sleep
          else:
            sleep_time = max(1 / self.get_plugin_loop_resolution() - it_time, 0.00001)
          # now check for previous errors
          sleep_time = self.__maybe_handle_exec_errors(sleep_time=sleep_time)
          # end error checking
          self.start_timer('loop_sleep')
          if sleep_time > 0:
            sleep(sleep_time)
          self.end_timer('loop_sleep')
        except:
          self._plugin_loop_in_exec = False
          SLEEP_ON_LOOP_ERROR = 10
          msg = "CRITICAL ERROR in business plugin loop (sleep {}) on: {}".format(
            SLEEP_ON_LOOP_ERROR, self.log.get_error_info()
          )
          info = traceback.format_exc()
          self.P(msg + '\n' + info, color='r')
          self._create_error_notification(msg=msg, displayed=True)
          sleep(SLEEP_ON_LOOP_ERROR)
          self.has_loop_error = True
          self.__plugin_loop_errors += 1 # increase nr of errors
          self.__post_error_ok_count += 3 # increase loop error level
        #end try
      #endwhile
    #endif plugin_loop_initialized

    try:
      self.__cleanup()
    except Exception as exc:
      info = traceback.format_exc()
      self.P(f"CRITICAL ERROR in cleanup: {exc}:\n{info}", color='r')

    # handle commands emitted on close 
    commands = self.get_commands_after_exec()
    if len(commands) > 0:
      self.commands_deque.append(commands)

    self.P(ct.PLUGIN_END_PREFIX + "* * * * Business Plugin thread {} execution ended * * * *".format(self.instance_hash), color='y')
    return


  
  def __cleanup(self):
    #### del self.__upstream_inputs_deque ### this should NOT be called here - please do not ADD/DECOMMENT
    self._on_close()
    return




import numpy as np
import subprocess as sp
import shlex
from time import sleep

import traceback



class FFmpegWriter(object):
  def __init__(self, 
               filename, 
               fps, 
               frameSize, 
               log, 
               method='h264', 
               overall_options=None, 
               use_shell=False,
               pipe_cmd=''):
    """
    

    Parameters
    ----------
    filename : str
      
    fps : int
      
    frameSize : tuple
        
    
    log : Logger
      
    
    method : str, optional
      'h264' or other encoding methods - use default. The default is 'h264'.
      
    overall_options : str, optional
      `-stats_period 10` will show from 10 to 10 seconds . The default is None.

    Returns
    -------
    None.

    """
    assert method in ['h264', 'h265']
    method = method[1:]
    self._codec = method
    self._fn = filename
    self._fps = fps
    self._fsize = frameSize
    self.log = log
    self._write_timer = 'h{}_write'.format(self._codec)
    width, height = frameSize
    
    if isinstance(overall_options, str) and len(overall_options) > 0:
      overall_options = overall_options + ' '
    else:
      overall_options = ''

    raw_cmd = f'ffmpeg -y {overall_options}-s {width}x{height} -pixel_format bgr24 -f rawvideo -r {fps} -i pipe: -video_track_timescale 30k -vcodec libx{method} -pix_fmt yuv420p -crf 24 "{filename}"'
    if use_shell and len(pipe_cmd.strip()) > 0:
      if pipe_cmd.strip()[0] != '|':
        pipe_cmd = '| ' + pipe_cmd
      raw_cmd = raw_cmd + ' ' + pipe_cmd
    
    self._cmd = raw_cmd    
    self.log.P("Running ffmpeg:\n{}".format(raw_cmd))
    starter = raw_cmd if use_shell else shlex.split(raw_cmd)
    
    # TODO (S): must get result output!
    self.process = sp.Popen(
      starter, 
      stdin=sp.PIPE,
      shell=use_shell,
    )
    sleep(2)
    # TODO (S): show & raise error if the process resulted in exec error (ie. wrong parameters)
    return
  
  @property
  def command(self):
    return self._cmd
  
  def get_command(self):
    return self.command
  
  def isOpened(self):
    return self.process.poll() is None
  
  def write(self, frame):
    assert isinstance(frame, np.ndarray), "write input must be ndarray"
    if len(frame.shape) == 4:
      b_size = frame.shape[0]
    else:
      b_size = 1
    # Write raw video frame to input stream of ffmpeg sub-process.
    self.log.start_timer(self._write_timer, section='FFmpegWriter')
    self.log.start_timer('bytes_b'+str(b_size), section='FFmpegWriter')
    buff = frame.tobytes()
    self.log.stop_timer('bytes_b'+str(b_size), section='FFmpegWriter')
    self.log.start_timer('write_b'+str(b_size), section='FFmpegWriter')
    try:
      self.process.stdin.write(buff)
    except Exception as e:
      raise ValueError("Error in FFmpegWriter: '{}'. Using:\n{}".format(
        str(e), self.command,
        )
      )
    self.log.stop_timer('write_b'+str(b_size), section='FFmpegWriter')
    self.log.stop_timer(self._write_timer, section='FFmpegWriter')
    return
  
  def release(self):
    # Close and flush stdin
    self.process.stdin.close()    
    # Wait for sub-process to finish
    self.process.wait()    
    # Terminate the sub-process
    self.process.terminate()
    return
  

      
      
  
  



#global dependencies
import numpy as np
from itertools import combinations

from shapely.geometry import Polygon, LineString

###
###START OBJECTS INTERSECTION
###


def np_vec_iou(boxes1, boxes2):
  """
  This method calculates iou between two sets of boxes
  """
  top, left, bottom, right = np.split(boxes1, 4, axis=1)
  _top, _left, _bottom, _right = np.split(boxes2, 4, axis=1)
  xA = np.maximum(left, np.transpose(_left))
  yA = np.maximum(top, np.transpose(_top))
  xB = np.minimum(right, np.transpose(_right))
  yB = np.minimum(bottom, np.transpose(_bottom))
  inter_length = (xB - xA + 1)
  inter_height = (yB - yA + 1)
  inter_area = np.maximum(inter_length, 0) * np.maximum(inter_height, 0)
  boxA_area = (right - left + 1) * (bottom - top + 1)
  boxB_area = (_right - _left + 1) * (_bottom - _top + 1)
  iou = inter_area / (boxA_area + np.transpose(boxB_area) - inter_area)
  return iou, (inter_area / boxA_area, inter_area / np.transpose(boxB_area)), (inter_length, inter_height)

def np_intersect_over_min_area(boxes1, boxes2):
  """
  This method calculates iou between two sets of boxes
  """
  top, left, bottom, right = np.split(boxes1, 4, axis=1)
  _top, _left, _bottom, _right = np.split(boxes2, 4, axis=1)
  xA = np.maximum(left, _left)
  yA = np.maximum(top, _top)
  xB = np.minimum(right, _right)
  yB = np.minimum(bottom, _bottom)
  inter_length = (xB - xA + 1)
  inter_height = (yB - yA + 1)
  inter_area = np.maximum(inter_length, 0) * np.maximum(inter_height, 0)
  boxA_area = (right - left + 1) * (bottom - top + 1)
  boxB_area = (_right - _left + 1) * (_bottom - _top + 1)
  # iou = inter_area / (boxA_area + np.transpose(boxB_area) - inter_area)
  return inter_area / np.minimum(boxA_area, boxB_area) #(inter_area / boxA_area, inter_area / np.transpose(boxB_area)), (inter_length, inter_height)

def intersection_boxes_box(boxes, box):
  """
  This method receives a list of bounding boxes and a single bounding box and
  computes for each box in the list the intersection area with the single box.
  """
  top, left, bottom, right = np.split(boxes, 4, axis=1)
  _top, _left, _bottom, _right = box

  xA = np.maximum(left, _left)
  yA = np.maximum(top, _top)
  xB = np.minimum(right, _right)
  yB = np.minimum(bottom, _bottom)

  inter_length = (xB - xA + 1)
  inter_height = (yB - yA + 1)
  inter_area = np.maximum(inter_length, 0) * np.maximum(inter_height, 0)
  return inter_area


def boxes_intersect_box(boxes, box):
  """
  This method receives a list of bounding boxes and a single bounding box and
  computes for each box in the list whether or it intersects the single box.
  """
  inter_area = intersection_boxes_box(boxes, box)
  return inter_area > 0


def boxes_areas(boxes):
  "This method receives a list of bounding boxes and computes the area for each of them"
  top, left, bottom, right = np.split(boxes, 4, axis=1)
  areas = (bottom - top + 1) * (right - left + 1)
  return areas


def iou_boxes_box(boxes, box):
  """
  This method receives a list of bounding boxes and a single bounding box and
  computes the IoU scores for each bounding box from boxes with the one in box
  """
  inter_area = intersection_boxes_box(boxes, box)
  areas = boxes_areas(boxes)
  t, l, b, r = box
  area = (b - t + 1) * (r - l + 1)

  iou = inter_area / (areas + area - inter_area)

  return iou


def count_unique_boxes_intersection(dict_boxes, direction=None):
  """
  This method returns the number of unique boxes that intersect. 
  Method expects `dict_boxes` to contain all the boxes that you need to verify for intersection between them.
  `dict_boxes` keys will be of type int or str
  `dict_boxes` values will contain [TOP, LEFT, BOTTOM, RIGHT] values specific to each box
    dict_boxes = {
      0: [10, 20, 30, 40],
      1: [50, 60, 70, 80],
      3: [10, 20, 40, 60]
      }  
  """
  
  _len = len(dict_boxes)
  if _len < 2: 
    return _len
  
  unique_boxes = set()
  combs = list(combinations(dict_boxes.keys(), 2))
  for comb in combs:
    TLBR1, TLBR2 = dict_boxes[comb[0]], dict_boxes[comb[1]]
    iou, _, (inter_length, inter_height) = np_vec_iou(
      boxes1=np.array([TLBR1]), 
      boxes2=np.array([TLBR2])
      )
    res = False
    if iou > 0:
      res = True
      if direction is not None and direction != 'BOTH':
        intersection_direction = 'VERTICAL' if inter_length > inter_height else 'HORIZONTAL'
        res = direction == intersection_direction
      #endif
    #endif
    if res:
      unique_boxes.update([comb[0], comb[1]])
  #endfor
  return len(unique_boxes) or 1


def boxes_are_intersecting(tlbr_box1, tlbr_box2):
  """
  This method verifies if two boxes to intersecting with each other based on their 
  [TOP, LEFT, BOTTOM, RIGHT] coordinates.
  """
  
  _top, _left, _bottom, _right = tlbr_box1
  top, left, bottom, right = tlbr_box2
  if _left > right or _right < left or _top > bottom or _bottom < top:
    return False
  return True


def intersect_box_irregular_target(box, target_area):
  """
  This method calculates the intersection percent between a `box` (TLBR area) and a
  polygon/non-regular area
  
  The `box` argument can be provided either in a simple TLBR list format or
  as a polygon format.
  
  The method returns the intersection percent between provided shapes.
  """
  
  if not isinstance(box[0], list):
    top    = box[0]
    left   = box[1]
    bottom = box[2]
    right  = box[3]
    box = [
      [left, top], 
      [right, top], 
      [right, bottom], 
      [left, bottom], 
      [left, top]
      ]
  elif len(box) == 4:# box[-1] != box[0]
    box.append(box[0])
  box = np.array(box).astype(np.float64)
  if target_area[0] == target_area[-1]:
    target_area_shapely = Polygon(target_area)
  else:
    target_area_shapely = LineString(target_area)
  box_shapely = Polygon(box)
  intersect = box_shapely.intersection(target_area_shapely)
  if box_shapely.area == 0:
    intersect_value = 0
  else:
    intersect_value = intersect.area / box_shapely.area
  return intersect_value
  
  
def intersect_box_regular_target(lst_boxes_tlbr, target_area):
  """
  This method calculates the intersection percent between a list of boxes
  described by their TLBR and a regular target area also decribed by TLBR.
  
  This method return the intersection percent for each box with the target area.
  """
  
  _, intersect, _ = np_vec_iou(lst_boxes_tlbr, np.array([target_area]))
  intersect = intersect[0].ravel()
  return intersect      



# def filter_boxes_irregular_target(lst_box, target_area, prc_intersect):
def keep_boxes_inside_irregular_target_area(lst_box, target_area, prc_intersect):
  """
  This method filter a list of boxes described by their TLBR values with a target area
  non-rectangular. 
  The filtering steps takes into account `prc_intersect` and only the boxes with intersection
  over the the `prc_intersect` will be kept.
  """
  
  l = [intersect_box_irregular_target(x['TLBR_POS'], target_area) for x in lst_box]
  arr_intersect = np.array(l)
  sel = arr_intersect >= prc_intersect
  intersect_values = arr_intersect[sel]
  lst_box = np.array(lst_box)[sel].tolist()
  return lst_box, intersect_values


# def filter_boxes_regular_target(lst_box, target_area, prc_intersect):
def keep_boxes_inside_regular_target_area(lst_box, target_area, prc_intersect):
  """
  This method filter a list of boxes described by their TLBR values with a target area
  also described by TLBR value
  The filtering steps takes into account `prc_intersect` and only the boxes with intersection
  over the the `prc_intersect` will be kept.
  """
  
  lst_boxes_tlbr = np.array([x['TLBR_POS'] for x in lst_box])
  intersect = intersect_box_regular_target(
    lst_boxes_tlbr=lst_boxes_tlbr,
    target_area=target_area
    )
  sel = intersect >= prc_intersect
  intersect_values = intersect[sel]
  lst_box = np.array(lst_box)[sel].tolist()
  return lst_box, intersect_values

def convert_points_to_tlbr(points):
  """
  This method converts points to TLBR coordinates
  """
  
  arr = np.array(points)
  left = arr[:, 0].min()
  right = arr[:, 0].max()
  top = arr[:, 1].min()
  bottom = arr[:, 1].max()
  return [top, left, bottom, right]

def convert_tlbr_to_points(tlbr):
  top, left, bottom, right = tlbr
  points = [[left, top], [right, top], [right, bottom], [left, bottom], [left, top]]
  return points

def keep_boxes_outside_irregular_target_area(lst_box, target_area):
  """
  This method keeps only boxes that are outside of a irregular target area. Boxes are described by their TLBR values.
  The filtering steps takes into account `prc_intersect` and only the boxes with intersection
  over the the `prc_intersect` will be kept.
  """
  
  l = [intersect_box_irregular_target(x['TLBR_POS'], target_area) for x in lst_box]
  arr_intersect = np.array(l)
  sel = arr_intersect == 0
  lst_box = np.array(lst_box)[sel].tolist()
  return lst_box


def keep_boxes_outside_regular_target_area(lst_box, target_area):
  """
  This method keeps only boxes that are outside of a regular target area. Boxes are described by their TLBR values with a target area also described by TLBR value
  The filtering steps takes into account `prc_intersect` and only the boxes with intersection
  over the the `prc_intersect` will be kept.
  """
  
  lst_boxes_tlbr = np.array([x['TLBR_POS'] for x in lst_box])
  intersect = intersect_box_regular_target(
    lst_boxes_tlbr=lst_boxes_tlbr,
    target_area=target_area
    )
  sel = intersect == 0
  lst_box = np.array(lst_box)[sel].tolist()
  return lst_box


def intersect(np_box1, np_box2):
  assert np_box1.shape[0] == np_box2.shape[0]
  np_top     = np_box1[:,0] <= np_box2[:,0]
  np_left    = np_box1[:,1] <= np_box2[:,1]
  np_bottom  = np_box1[:,2] >= np_box2[:,2]
  np_right   = np_box1[:,3] >= np_box2[:,3]
  result = np.all([np_top, np_left, np_bottom, np_right], axis=0)
  return result


def get_boxes_overlapp_relations(np_boxes1, np_boxes2):
  """
  This method returns the overlapp mapping between two list of boxes.

  Parameters
  ----------
  np_boxes1 : np.ndarray
    Array of TLBR boxes. (ex: persons)
  np_boxes2 : np.ndarray
    Array of TLBR boxes. (ex: faces)

  Returns
  -------
  rel_boxes1 : dictionary
    Each box1 item will be maped to a list of box2 items that do intersect with the box1 item
    -> keys will be formed by boxes1 ids
    -> values will be formed by lists of ids from boxes2 that are intersected with the current box2
  rel_boxes2 : dictionary
    Each box2 item will be maped to a list of box1 items that do intersect with the box2 item
    -> keys will be formed by boxes2 ids
    -> values will be formed by lists of ids from boxes1 that are intersected with the current box2

  """
  nr_boxes1 = np_boxes1.shape[0]
  nr_boxes2 = np_boxes2.shape[0]
  
  if nr_boxes1 == 0 or nr_boxes2 == 0:
    rel_boxes1 = {i:[] for i in range(nr_boxes1)}
    rel_boxes2 = {i:[] for i in range(nr_boxes2)}
  else:
    np_boxes1 = np.repeat(np_boxes1, repeats=nr_boxes2, axis=0) #repeat each box1 item to compare it with each box2 item. shape = (nr_boxes1*nr_boxes2, N)
    np_boxes2 = np.vstack([np_boxes2] * nr_boxes1) #for each item in box1, create a replica of box2 items. shape = (nr_boxes2*nr_boxes1, N)
    
    np_overlapps = intersect(np_boxes1, np_boxes2)  #check overlapping boxes. shape = (nr_boxes1*nr_boxes2,)
    
    #create overlapp relations between box1 items and box2 items
    np_overlapps1 = np_overlapps.reshape(nr_boxes1, nr_boxes2) #reshape elements such that each box1 item contains box2 items. shape = (nr_boxes1, nr_boxes2)
    rel_boxes1 = {i: np.argwhere(x).ravel().tolist() for i,x in enumerate(np_overlapps1)} #key represented by box1 item id, value represented by list of box2 ids that overlapp over current box1 item
    
    #create overlapp relations between box1 items and box2 items
    np_overlapps2 = np_overlapps1.T ##reshape such that each box2 item gets its corresponding box1 items. shape = (nr_boxes2, nr_boxes1)
    rel_boxes2 = {i: np.argwhere(x).ravel().tolist() for i,x in enumerate(np_overlapps2)} #key represented by box2 item id, value represented by list of box1 ids that overlapp over current box2 item
  return rel_boxes1, rel_boxes2


def get_non_overlappping_boxes(np_boxes1, np_boxes2):
  """
  This method analyzes two ndarray of TLBR boxes and return only the boxes that are non-overlapping

  Parameters
  ----------
  np_boxes1 : np.ndarray
    Array of TLBR boxes (ex: persons)
  np_boxes2 : np.ndarray
    Array of TLBR boxes (ex: faces)

  Returns
  -------
  np_box1 : np.ndarray
    Only the initial box1 items that are non overlapping with box2 items
  np_box2 : np.ndarray
    Only the initial box2 items that are non overlapping with box1 items

  """
  
  rel_boxes1, rel_boxes2 = get_boxes_overlapp_relations(np_boxes1, np_boxes2)
  
  #extract boxes1 that are not overlapping
  sel_box1 = [k for k,v in rel_boxes1.items() if not v]
  np_box1 = np_boxes1[sel_box1]
  
  #extract boxes2 that are not overlapping
  sel_box2 = [k for k,v in rel_boxes2.items() if not v]
  np_box2 = np_boxes2[sel_box2]
    
  return np_box1, np_box2


def get_only_overlapping_boxes(np_boxes1, np_boxes2):
  """
  This method analyzes two arrays of TLBR boxes and returns only boxes that are intersecting.

  Parameters
  ----------
  np_boxes1 : np.ndarray
    Array of TLBR boxes (ex: persons)
  np_boxes2 : np.ndarray
    Array of TLBR boxes (ex: faces)

  Returns
  -------
  np_box1 : np.ndarray
    Only the initial box1 items that are overlapping with box2 items
  np_box2 : np.ndarray
    Only the initial box2 items that are overlapping with box1 items

  """
  rel_boxes1, rel_boxes2 = get_boxes_overlapp_relations(np_boxes1, np_boxes2)
  
  #extract boxes1 that are overlapping
  sel_box1 = [k for k,v in rel_boxes1.items() if v]
  np_box1 = np_boxes1[sel_box1]
  
  #extract boxes2 that are not overlapping
  sel_box2 = [k for k,v in rel_boxes2.items() if v]
  np_box2 = np_boxes2[sel_box2]
    
  return np_box1, np_box2


def unify_overlapping_boxes(np_boxes1, np_boxes2, box1_priority_on_overlap=True):
  """
  This method analyzes two arrays of TLBR boxes and returns the "union" of those two arrays, based on choosen box priority.
  For example: if `np_boxes1` are persons, `np_boxes2` are faces and `box1_priority_on_overlap` is True, then the method will:
    - keep & return all persons that are non intersecting with a face box
    - keep & return all faces that are non intersecting with a person box
    - keep person in case that person is intersecting a face box
  

  Parameters
  ----------
  np_boxes1 : np.ndarray
    Array of TLBR boxes (ex: persons)
  np_boxes2 : np.ndarray
    Array of TLBR boxes (ex: faces)
  box1_priority_on_overlap : TYPE, optional
    Choose he box that will be kept in case of intersection

  Returns
  -------
  np_box1 : np.ndarray
    Will contain all non-intersecting box1 items. In case `box1_priority_on_overlap` is set to True, will also contain box1 items that do intersect with box2 items
  np_box2 : np.ndarray
    Will contain all non-intersecting box2 items. In case `box1_priority_on_overlap` is set to False, will also contain box2 items that do intersect with box1 items

  """
  rel_boxes1, rel_boxes2 = get_boxes_overlapp_relations(np_boxes1, np_boxes2)
  
  #extract boxes1 that are not overlapping
  sel_box1 = [k for k,v in rel_boxes1.items() if not v]
  
  #extract boxes2 that are not overlapping
  sel_box2 = [k for k,v in rel_boxes2.items() if not v]
  
  if box1_priority_on_overlap:
    sel_box1+= [k for k,v in rel_boxes1.items() if v]
    
  else:
    sel_box2+= [k for k,v in rel_boxes2.items() if v]
  
  np_box1 = np_boxes1[sel_box1]
  np_box2 = np_boxes2[sel_box2]
  
  return np_box1, np_box2

###
###STOP OBJECTS INTERSECTION
###



import sys

class _PackageLoaderMixin(object):
  """
  Mixin for package loader functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_PackageLoaderMixin, self).__init__()

  @staticmethod
  def package_loader(package_name, as_bool=True, return_package=False):
    """
    returns True (or nr of similar loaded packages) for a certain package
    if `return_package` == True then returns a reference to the module
    """
    i_res = sum([package_name in x for x in list(sys.modules.keys())])
    if return_package:
      if i_res > 0:
        return sys.modules[package_name]
      else:
        return None
    else:
      if as_bool:
        return i_res > 0
      else:
        return i_res

  def reset_seeds(self, seed=123, packages=['np', 'rn', 'tf', 'th']):
    """
    this method resets all possible random seeds in order to ensure
    reproducible results
    this method resets for:
        numpy, random, tensorflow, torch
    """
    _np, _rn, _tf, _th = None, None, None, None

    if 'np' in packages:
      _np = self.package_loader('numpy', return_package=True)

    if 'rn' in packages:
      _rn = self.package_loader('random', return_package=True)

    if 'tf' in packages:
      if 'tensorflow' in sys.modules:
        _tf = self.package_loader('tensorflow', return_package=True)

    if 'th' in packages:
      if 'torch' in sys.modules:
        _th = self.package_loader('torch', return_package=True)

    if _np is not None:
      self.P("Setting random seed {} for 'Numpy'".format(seed))
      _np.random.seed(seed)

    if _rn is not None:
      self.P("Setting random seed {} for 'random'".format(seed))
      _rn.seed(seed)

    if _tf is not None:
      self.P("Setting random seed {} for 'tensorflow'".format(seed))
      if _tf.__version__[0] == '2':
        _tf.random.set_seed(seed)
      else:
        _tf.set_random_seed(seed)

    if _th is not None:
      self.P("Setting random seed {} for 'torch'".format(seed))
      _th.manual_seed(seed)
    return



from naeural_core.business.base.network_processor import NetworkProcessorPlugin

_CONFIG = {
  **NetworkProcessorPlugin.CONFIG,

  'NUMBER_OF_PAYLOADS': 5,
  'PAYLOAD_PERIOD': 5,
  'ACCEPT_SELF': True,
  "ALLOW_EMPTY_INPUTS": True,
  'CONVERT_TO_STR': False,

  'VALIDATION_RULES' : {
    **NetworkProcessorPlugin.CONFIG['VALIDATION_RULES'],
  },
}


class NetworkListenerDebugPlugin(NetworkProcessorPlugin):
  _CONFIG = _CONFIG
  def get_payload_dict(self):
    faulty_stuff = {
      0: {'something': 'wrong', 'number': 22, 'sss': [222, 'ppp']},
      4: {'something': 'wrong', 'number': 22, 'sss': [222, 'ppp']},
      5: {'something': 'wrong', 'number': 22, 'sss': [222, 'ppp']},
      6: {'something': 'wrong', 'number': 22, 'sss': [222, 'ppp']},
      3333: {'something': 'wrong', 'number': 22, 'sss': [222, 'ppp']},
    }
    if self.cfg_convert_to_str:
      faulty_stuff = {str(k): str(v) for k, v in faulty_stuff.items()}
    return {
      'AA_network_debug_counter': self.total_payload_count,
      'AA_network_debug_sender': self.ee_id,
      'AA_network_debug_sender_full': '|'.join([self.ee_id, self._stream_id, self.get_instance_id()]),
      'AA_network_debug_path': (self.ee_id, self.ee_addr, self._stream_id, self.get_instance_id()),
      'maybe_faulty_data': faulty_stuff
    }

  @NetworkProcessorPlugin.payload_handler
  def on_payload_debug(self, data: dict):
    payload_number = data.get('AA_network_debug_counter', -1)
    payload_sender = data.get('AA_network_debug_sender', 'unknown')
    payload_sender_full = data.get('AA_network_debug_sender_full', 'unknown')
    self.P(f'Payload number {payload_number} received from {payload_sender_full}.')
    return

  def process(self):
    if self.time() - self.last_payload_time > self.cfg_payload_period:
      self.P(f'Adding {self.cfg_number_of_payloads} payloads.')
      for i in range(self.cfg_number_of_payloads):
        self.add_payload_by_fields(**self.get_payload_dict())
      # endfor each payload
    # endif time to add payloads
    return



from naeural_core.business.default.web_app.fast_api_web_app import FastApiWebAppPlugin as BasePlugin

__VER__ = '0.1.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,
  
  "RUNS_ONLY_ON_SUPERVISOR_NODE" : True,

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },
}


class SupervisorFastApiWebApp(BasePlugin):
  CONFIG = _CONFIG

  def on_init(self):
    self.__supervisor_fastapi_plugin_running = None
    super(SupervisorFastApiWebApp, self).on_init()
    return
  
  def __get_current_epoch(self):
    """
    Get the current epoch of the node.

    Returns
    -------
    int
        The current epoch of the node.
    """    
    return self.netmon.epoch_manager.get_current_epoch()
    
  
  def __sign(self, data):
    """
    Sign the given data using the blockchain engine.
    Returns the signature. 
    Use the data param as it will be modified in place.
    """
    signature = self.bc.sign(data, add_data=True, use_digest=True)
    return signature

  def _get_response(self, dct_data: dict):
    """
    
    Create a response dictionary with the given data.

    Parameters
    ----------
    dct_data : dict
        The data to include in the response - data already prepared 

    Returns
    -------
    dict

    """
    try:
      str_utc_date = self.datetime.now(self.timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
      dct_data['server_info'] = {
        'alias': self.node_id,
        'version': self.ee_ver,
        'time': str_utc_date,
        'current_epoch': self.__get_current_epoch(),
        'uptime': str(self.timedelta(seconds=int(self.time_alive))),
      }
      self.__sign(dct_data) # add the signature over full data
    except Exception as e:
      self.P("Error in `get_response`: {}".format(e), color='r')
      preexisting_error = dct_data.get('error', "")
      dct_data['error'] = f"{preexisting_error} - {e}"
    return dct_data  
  

  @property
  def __is_enabled(self):
    res = not self.cfg_disabled and self.cfg_ngrok_edge_label is not None and self.is_supervisor_node
    if res != self.__supervisor_fastapi_plugin_running:
      self.__supervisor_fastapi_plugin_running = res
      if res:
        self.P(f"{self.__class__.__name__} is enabled")
      else:
        disabled_cause = 'disabled by config parameter' if self.cfg_disabled \
          else 'no ngrok edge label provided' if self.cfg_ngrok_edge_label is None \
          else 'not a supervisor node'
        msg = f"{self.__class__.__name__} is disabled. (cause: {disabled_cause})"
        self.P(msg, color='r', boxed=True)
    # endif changed state
    return res

  def _process(self):
    if self.cfg_runs_only_on_supervisor_node and not self.__is_enabled:
      return None
    return super(SupervisorFastApiWebApp, self)._process()



# local dependencies
from naeural_core.serving.default_inference.th_rface import ThRface as ParentServingProcess

_CONFIG = {
  **ParentServingProcess.CONFIG,

  'IMAGE_HW': (640, 640),  # The model is size-invariant, but is trained on this size ( 1080, 1920),#
  'MAX_BATCH_FIRST_STAGE': 5,
  "MODEL_WEIGHTS_FILENAME": "20230723_RET_FACE_MOBILE_bs1_nms.ths",
  "URL": "minio:retina_face/20230723_RET_FACE_MOBILE_bs1_nms.ths",

  'VALIDATION_RULES': {
    **ParentServingProcess.CONFIG['VALIDATION_RULES'],
  },
}

DEBUG_COVERED_SERVERS = False


class ThRfaceS(ParentServingProcess):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(ThRfaceS, self).__init__(**kwargs)
    return




class _BasePluginAPIMixin:
  def __init__(self) -> None:
    super(_BasePluginAPIMixin, self).__init__()
    
    self.__chain_state_initialized = False
    return
  
  # Obsolete
  def _pre_process(self):
    """
    Called before process. Currently (partially) obsolete

    Returns
    -------
    TBD.

    """
    return
  
  def _post_process(self):
    """
    Called after process. Currently (partially) obsolete

    Returns
    -------
    TBD.

    """
    return
  
  
  def step(self):
    """
    The main code of the plugin (loop iteration code). Called at each iteration of the plugin loop.

    Returns
    -------
    None.

    """
    return
  
  
  def process(self):
    """
    The main code of the plugin (loop iteration code). Called at each iteration of the plugin loop.

    Returns
    -------
    Payload.

    """
    return self.step()
  
  def _process(self):
    """
    The main code of the plugin (loop iteration code.

    Returns
    -------
    Payload.

    """
    return self.process()

  
  def on_init(self):
    """
    Called at init time in the plugin thread.

    Returns
    -------
    None.

    """      
    return
  
  def _on_init(self):
    """
    Called at init time in the plugin thread.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_init` called for plugin initialization...")
    self.on_init()
    return


  def on_close(self):
    """
    Called at shutdown time in the plugin thread.

    Returns
    -------
    None.

    """      
    return


  def _on_close(self):
    """
    Called at shutdown time in the plugin thread.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_close` called for plugin cleanup at shutdown...")
    self.maybe_archive_upload_last_files()
    self.on_close()
    return

  def on_command(self, data, **kwargs):
    """
    Called when a INSTANCE_COMMAND is received by the plugin instance.
    
    The command is sent via `cmdapi_send_instance_command` as in below simplified example:
    
    ```python
      pipeline = "some_app_pipeline"
      signature = "CONTAINER_APP_RUNNER"
      instance_id = "CONTAINER_APP_1e8dac"
      node_address = "0xai_1asdfG11sammamssdjjaggxffaffaheASSsa"
      
      instance_command = "RESTART"
      
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=node_address,
      )
    ```
    
    while the `on_command` method should look like this:
    
    ```python
      def on_command(self, data, **kwargs):
        if data == "RESTART":
          self.P("Restarting ...")
        elif data == "STOP":
          self.P("Stopping ...")
        else:
          self.P(f"Unknown command: {data}")
        return
    ```
    
    The `instance_command` is passed to this method as `data` and in fact can be a dict with extra data. If
    `instance_command` contains `COMMAND_PARAMS` dict then all the key-value pairs in the `COMMAND_PARAM` dict 
    will be passed as kwargs to this method - see below example:
    
    ```python
      instance_command = {
        "COMMAND_PARAMS": {
          "some_param1": "value1",
          "some_param2": "value2",
        }
      }
      
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=node_address,
      )
    ```
    Then this `on_command` method should look like this:
    
    ```python
    
      def on_command(self, data, some_param1=None, some_param2=None, **kwargs):
        if some_param1:
          self.P(f"Received some_param1: {some_param1}")
        if some_param2:
          self.P(f"Received some_param2: {some_param2}")
        # Process the command here
        return
    ```

    Parameters
    ----------
    data : any
      object, string, etc.

    Returns
    -------
    None.

    """
    return

  def _on_command(self, data, default_configuration=None, current_configuration=None, **kwargs):
    """
    Called when a INSTANCE_COMMAND is received by the plugin instance.
    
    The command is sent via `cmdapi_send_instance_command` as in below simplified example:
    
    ```python
      pipeline = "some_app_pipeline"
      signature = "CONTAINER_APP_RUNNER"
      instance_id = "CONTAINER_APP_1e8dac"
      node_address = "0xai_1asdfG11sammamssdjjaggxffaffaheASSsa"
      
      instance_command = "RESTART"
      
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=node_address,
      )
    ```
    
    while the `on_command` method should look like this:
    
    ```python
      def on_command(self, data, **kwargs):
        if data == "RESTART":
          self.P("Restarting ...")
        elif data == "STOP":
          self.P("Stopping ...")
        else:
          self.P(f"Unknown command: {data}")
        return
    ```
    
    The `instance_command` is passed to this method as `data` and in fact can be a dict with extra data. If
    `instance_command` contains `COMMAND_PARAMS` dict then all the key-value pairs in the `COMMAND_PARAM` dict 
    will be passed as kwargs to this method - see below example:
    
    ```python
      instance_command = {
        "COMMAND_PARAMS": {
          "some_param1": "value1",
          "some_param2": "value2",
        }
      }
      
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=node_address,
      )
    ```
    Then this `on_command` method should look like this:
    
    ```python
    
      def on_command(self, data, some_param1=None, some_param2=None, **kwargs):
        if some_param1:
          self.P(f"Received some_param1: {some_param1}")
        if some_param2:
          self.P(f"Received some_param2: {some_param2}")
        # Process the command here
        return
    ```
    
    Parameters
    ----------
    data : any
      object, string, etc.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_command`...")

    if (isinstance(data, str) and data.upper() == 'DEFAULT_CONFIGURATION') or default_configuration:
      self.P("Received \"DEFAULT_CONFIGURATION\" command...")
      self.add_payload_by_fields(
        default_configuration=self._default_config,
        command_params=data,
      )
      return
    if (isinstance(data, str) and data.upper() == 'CURRENT_CONFIGURATION') or current_configuration:
      self.P("Received \"CURRENT_CONFIGURATION\" command...")
      self.add_payload_by_fields(
        current_configuration=self._upstream_config,
        command_params=data,
      )
      return

    self.on_command(data, **kwargs)
    return


  def _on_config(self):
    """
    Called when the instance has just been reconfigured

    Parameters
    ----------
    None

    Returns
    -------
    None.

    """
    self.P("Default plugin {} `_on_config` called...".format(self.__class__.__name__))
    if hasattr(self, 'on_config'):
      self.on_config()
    return


  ###
  ### Chain State
  ### 
  
  def __chainstorage_memory(self):
    return self.plugins_shmem
  
  def __maybe_wait_for_chain_state_init(self):
    # TODO: raise exception if not found after a while

    while not self.plugins_shmem.get('__chain_storage_set'):
      self.sleep(0.1)
    
    if not self.__chain_state_initialized:
      self.P(" ==== Chain state initialized.")
    self.__chain_state_initialized = True
    return
  
  def chainstore_set(self, key, value, readonly=False, token=None, debug=False):
    """
    Set data in the R1 Chain Storage
    
    Parameters
    ----------
    key : str
      Key
      
    value : any
      Value
      
    readonly : bool, optional
      Read-only flag, by default False
      
    token : any, optional
      Token, by default None
      
    """
    memory = self.__chainstorage_memory()
    result = False
    try:
      self.__maybe_wait_for_chain_state_init()
      func = memory.get('__chain_storage_set')
      specific_peers = self.cfg_chainstore_peers
      if func is not None:
        if debug:
          self.P("Setting data: {} -> {}".format(key, value), color="green")
        self.start_timer("chainstore_set")
        result = func(
          key, value, 
          readonly=readonly, token=token, peers=specific_peers, debug=debug
        )
        elapsed = self.end_timer("chainstore_set")        
        if debug:
          self.P(" ====> `chainstore_set` elapsed time: {:.6f}".format(elapsed), color="green")
      else:
        if debug:
          self.P("No chain storage set function found", color="red")
    except Exception as ex:
      self.P("Error in chainstore_set: {}".format(ex), color="red")
    return result
  
  
  def chainstore_get(self, key, token=None, debug=False):
    """
    Get data from the R1 Chain Storage
    
    Parameters
    ----------
    key : str
      Key
      
    token : any, optional
      Token, by default None
    """
    memory = self.__chainstorage_memory()
    self.__maybe_wait_for_chain_state_init()
    value = None
    msg = ""
    try:
      start_search = self.time()
      found = True
      while memory.get('__chain_storage_get') is None:
        self.sleep(0.1)
        if self.time() - start_search > 10:
          msg = "Error: chain storage get function not found after 10 seconds"
          self.P(msg, color="red")
          found = False
          break
      func = memory.get('__chain_storage_get')
      if func is not None:
        value = func(key, token=token, debug=debug)
        if debug:
          self.P("Getting data: {} -> {}".format(key, value), color="green")
      else:
        if debug:
          self.P("No chain storage get function found", color="red")
    except Exception as ex:
      msg = "Error in chainstore_get: {}".format(ex)
      self.P("Error in chainstore_get: {}".format(ex), color="red")
    return value

  
  def __hset_index(self, hkey):
    hkey_hash = self.get_hash(hkey, algorithm='sha256', length=10)
    return f"hs:{hkey_hash}:"

  
  def __hset_key(self, hkey, key):
    b64key = self.str_to_base64(key, url_safe=True)
    return self.__hset_index(hkey) + b64key

  
  def chainstore_hget(self, hkey, key, token=None, debug=False):
    """    
    This is a basic implementation of a hash get operation in the chain storage.
    It uses a hash-based string composition to create a composed key.
    """
    composed_key = self.__hset_key(hkey, key)
    return self.chainstore_get(composed_key, token=token, debug=debug)
  

  def chainstore_hset(self, hkey, key, value, readonly=False, token=None, debug=False):
    """
    This is a basic implementation of a hash set operation in the chain storage.
    It uses a hash-based string composition to create a composed key.
    """
    composed_key = self.__hset_key(hkey, key)
    return self.chainstore_set(composed_key, value, readonly=readonly, token=token, debug=debug)

  
  def chainstore_hlist(self, hkey : str, token=None, debug=False):
    index = self.__hset_index(hkey)
    memory = self.__chainstorage_memory()
    self.__maybe_wait_for_chain_state_init()
    chain_storage = memory.get('__chain_storage')
    result = []
    for key in chain_storage:
      if key.startswith(index):
        b64field = key[len(index):]
        field = self.base64_to_str(b64field, url_safe=True)
        result.append(field)
      #end if 
    #end for
    return result

  
  def chainstore_hkeys(self, hkey : str, token=None, debug=False):
    return self.chainstore_hlist(hkey, token=token, debug=debug)


  def chainstore_hgetall(self, hkey : str, token=None, debug=False):
    keys = self.chainstore_hlist(hkey, token=token, debug=debug)
    result = {}
    for key in keys:
      value = self.chainstore_hget(hkey, key, token=token, debug=debug)
      result[key] = value
    return result
    
  
  # # @property
  # # This CANNOT be a property, as it can be a blocking operation.
  # def _chainstorage(self): # TODO: hide/move/protect this
  #   self.__maybe_wait_for_chain_state_init()
  #   return self.plugins_shmem.get('__chain_storage')

  
  def get_instance_path(self):
    return [self.ee_id, self._stream_id, self._signature, self.cfg_instance_id]  
  
  ###
  ### END Chain State
  ###
  
    


import os

import torch as th
import json

from typing import Any, Mapping, Tuple, Sequence, Union

def create_from_torch(
  model: Union[th.nn.Module, th.jit.ScriptModule, th.jit.ScriptFunction],
  device : th.device,
  path : str,
  half : bool,
  input_names : Sequence[str],
  output_names : Sequence[str],
  args : Union[Tuple[Any, ...], th.Tensor],
  batch_axes : Mapping[str, Sequence[int]] | None = None,
  aggressive_shape_inference : bool = False,
  metadata = None
):
  """
  Export a torch or torchscript(this does not work for the moment
  because of a bug in pytorch) model to a onnx model on disk.
  Input and output names must be valid python identifiers.

  Parameters:
    model - the torch model
    path  - string, path on disk where the model is saved
    half - bool, export the model to fp16
    input_names - list of str, names to assign to inputs
    output_names - list of str, names to assign to outputs
    args - Can have one of 3 formats:
      - tensor
      - tuple of tensors
      - tuple of tensors except the last element which can be a dict
      Note that this matches the format of args in th.onnx.export
    batch_axes - Dictionary to specify the batch axes. Keys are input and
      output names. Values are a list of dynamic dimensions for the tensor.
      If None the first dimension of all input and output tensors is
      considered dynamic. Defaults to None.
    aggressive_shape_inference - Bool, run aggressive shape inference.
      Defaults to False.
    metadata - dict, will be added as metadata

  Returns: bool, True if successful, False otherwise.
  """

  import onnxsim
  import onnx
  from google.protobuf.json_format import MessageToDict

  if not isinstance(args, tuple):
    args = (args,)
  # Split arguments into positional and keyword arguments for model forward.
  exec_args = args
  exec_kwargs = {}
  if isinstance(args, tuple):
    if len(args) > 0 and isinstance(args[-1], dict):
      exec_kwargs = args[-1]
      exec_args = args[:-1]
    #endif last item is dict
  #endif tuple args

  # Make sure all used names are valid python identifiers,
  # otherwise we will have issues at inference time.
  for arg in input_names + output_names:
    if not isinstance(arg, str):
      return False
    if not arg.isidentifier():
      return False
  #endfor check names

  # Convert the model to the required precision and prepare it for
  # tracing.
  model.eval()
  model.to(device=device)
  if half:
    model = model.half()
  else:
    model = model.float()

  with th.no_grad():
    # Execute once otherwise we get trace-related issues.
    _ = model(*exec_args, **exec_kwargs)

  status = True
  try:
    if batch_axes is None:
      batch_axes = {}
      for name in input_names + output_names:
        batch_axes[name] = {0 : 'batch_size' }
    #endif init batch_axes
    with th.no_grad():
      th.onnx.export(
        model,
        args,
        path,
        verbose=False,
        do_constant_folding=True,
        input_names=input_names,
        output_names=output_names,
        dynamic_axes=batch_axes # Set batch as a dynamic axis
      )
    model_onnx = onnx.load(path)  # load onnx model
    # Make sure we have all inputs and outputs.
    if len(model_onnx.graph.input) != len(input_names) or len(model_onnx.graph.output) != len(output_names):
      raise ValueError('Not all input and output names specified')

    # This is essentially an optimization pipeline.
    # Do a simplification first, then run shape inference, then do more
    # simplifications.
    model_onnx, check = onnxsim.simplify(model_onnx) # simplify onnx model
    if aggressive_shape_inference:
      model_onnx = onnx.shape_inference.infer_shapes(model_onnx, True, False, True)
      model_onnx, check = onnxsim.simplify(model_onnx) # simplify onnx model
    assert check, 'Simplified ONNX model could not be validated'
    # Make sure the model is still valid.
    onnx.checker.check_model(model_onnx)  # check onnx model

    # There should be only one dynamic dimension.
    # If there is more than one either the user got the dynamic axes wrong
    # or there was an issue with shape inference.
    dyn_dims = set()
    for output in model_onnx.graph.output:
      output_dict = MessageToDict(output)
      type = output_dict.get('type')
      if type is None:
        continue
      tensor_type = type.get('tensorType')
      if tensor_type is None:
        continue
      shape = tensor_type.get('shape')
      if shape is None:
        continue
      dims = shape.get('dim')
      if dims is None:
        continue
      for dim in dims:
        param = dim.get('dimParam')
        if param is None:
          continue
        dyn_dims.add(param)
      #endfor all dims
    #endfor all outputs

    if len(dyn_dims) > 1:
      raise Exception("Error, multiple dynamic dimensions (perhaps shape inference issue?)")

    if metadata is None:
      metadata = {}

    # Set various metadata required at runtime.
    if metadata.get('precision') is None:
      if half:
        metadata['precision'] = 'fp16'
      else:
        metadata['precision'] = 'fp32'
    #endif precision metadata
    if metadata.get('input_names') is None:
      metadata['input_names'] = input_names
    if metadata.get('output_names') is None:
      metadata['output_names'] = output_names

    m1 = model_onnx.metadata_props.add()
    m1.key = 'onnx_metadata'
    m1.value = json.dumps(metadata)

    # Finally save the onnx model to disk.
    onnx.save(model_onnx, path)
  except Exception as e:
    print(str(e))
    os.unlink(path)
    status = False

  return status



from multiprocessing import Process, Pipe, Queue, JoinableQueue, Manager
import time
import sys
import gc
import numpy as np

counts = [
    10**3, 
    10**4, 10**5
]
def reset_global_msg_stats_dict(manager):
    return {
        'pipe': {
            key: manager.dict()
            for key in counts
        },
        'queue': {
            key: manager.dict()
            for key in counts
        }
    }


def reader_proc_q(queue, stats_dict):
    ## Read from the queue; this will be spawned as a separate Process
    while True:
        rcv_msg = queue.get()         # Read from the queue and do nothing
        if (isinstance(rcv_msg, str) and rcv_msg == 'DONE'):
            break
        else:
            stats_dict[rcv_msg.min()] = 2  # received

def reader_proc_jq(queue):
    ## Read from the queue; this will be spawned as a separate Process
    while True:
        rcv_msg = queue.get()         # Read from the queue and do nothing
        queue.task_done()
        if (isinstance(rcv_msg, str) and rcv_msg == 'DONE'):
            break
        del rcv_msg
        gc.collect()

def writer_q(count, queue, msg, stats_dict):
    ## Write to the queue
    for ii in range(0, count):
        stats_dict[ii] = 1  # sent
        queue.put(msg * 0 + ii)             # Write 'count' numbers into the queue
    queue.put('DONE')


def reader_proc_p(pipe, stats_dict):
    ## Read from the pipe; this will be spawned as a separate Process
    p_output, p_input = pipe
    p_input.close()    # We are only reading
    while True:
        rcv_msg = p_output.recv()    # Read from the output pipe and do nothing
        if isinstance(rcv_msg, str) and rcv_msg=='DONE':
            break
        else:
            stats_dict[rcv_msg.min()] = 2  # received

def writer_p(count, p_input, msg, stats_dict):
    for ii in range(0, count):
        stats_dict[ii] = 1  # sent
        p_input.send(msg * 0 + ii)             # Write 'count' numbers into the input pipe
    p_input.send('DONE')

def test_pipe(res_dct, counts, msg, msg_stats):
    for count in counts:
        # Pipes are unidirectional with two endpoints:  p_input ------> p_output
        p_output, p_input = Pipe()  # writer() writes to p_input from _this_ process
        reader_p = Process(target=reader_proc_p, args=((p_output, p_input), msg_stats['pipe'][count]))
        reader_p.daemon = True
        reader_p.start()     # Launch the reader process

        p_output.close()       # We no longer need this part of the Pipe()
        _start = time.time()
        writer_p(count, p_input, msg, msg_stats['pipe'][count]) # Send a lot of stuff to reader_proc()
        p_input.close()
        reader_p.join()
        elapsed = time.time() - _start
        res_dct['pipe'][count] = elapsed
        print("Sending {0} numbers to Pipe() took {1} seconds".format(count,
            elapsed))
    # endfor
    return

def test_queue(res_dct, counts, msg, msg_stats):
    pqueue = Queue(2) # writer() writes to pqueue from _this_ process
    for count in counts:
        ### reader_proc() reads from pqueue as a separate process
        reader = Process(target=reader_proc_q, args=((pqueue), msg_stats['queue'][count]))
        reader.daemon = True
        reader.start()        # Launch reader_proc() as a separate python process

        _start = time.time()
        writer_q(count, pqueue, msg, msg_stats['queue'][count])    # Send a lot of stuff to reader()
        reader.join()         # Wait for the reader to finish
        elapsed = time.time() - _start
        res_dct['queue'][count] = elapsed
        print("Sending {0} numbers to Queue() took {1} seconds".format(count, 
            (elapsed)))
    # endfor
    return

def test_joinable_queue(res_dct, counts, msg):
    pqueue = JoinableQueue(10) # writer() writes to pqueue from _this_ process
    for count in counts:             
        ### reader_proc() reads from pqueue as a separate process
        reader = Process(target=reader_proc_jq, args=((pqueue),))
        reader.daemon = True
        reader.start()        # Launch reader_proc() as a separate python process

        _start = time.time()
        writer_q(count, pqueue, msg)    # Send a lot of stuff to reader()
        reader.join()         # Wait for the reader to finish
        elapsed = time.time() - _start
        res_dct['queue'][count] = elapsed
        print("Sending {0} numbers to JoinableQueue() took {1} seconds".format(count, 
            (elapsed)))
    # endfor
    return

def print_aggregated(res_dct, counts, msg, msg_stats):
    global msg_stats_dict
    print(f'For msg of shape {msg.shape}:')
    for count in counts:
        missed_messages = {
            key: list(msg_stats[key][count].values()).count(1)
            for key in msg_stats.keys()
        }
        received_messages = {
            key: list(msg_stats[key][count].values()).count(2)
            for key in msg_stats.keys()
        }
        for key in msg_stats_dict.keys():
            print(f'[{key} and count={count}] Missed: {missed_messages[key]} | Received: {received_messages[key]}')
        res_dct['delta'][count] = res_dct['pipe'][count] - res_dct['queue'][count]
        print(f"Time delta for {count} messages: {results['delta'][count]}[Ratio: {results['delta'][count] / results['pipe'][count]}]")
    # endfor
    return

if __name__=='__main__':
    results = {
        'pipe': {},
        'queue': {},
        'delta': {}
    }
    manager = Manager()
    msg_stats_dict = manager.dict()
    msg_stats_dict = reset_global_msg_stats_dict(manager)

    # counts = [10**4, 10**5, 10**6, 10**7]
    # msg = np.random.randint(low=0, high=255, size=(10, 10, 10), dtype=np.uint8)

    msg = np.random.randint(low=0, high=255, size=(10, 10, 10))
    test_pipe(results, counts, msg, msg_stats_dict)
    gc.collect()
    test_queue(results, counts, msg, msg_stats_dict)
    # test_joinable_queue(results, counts, msg)
    gc.collect()
    print_aggregated(results, counts, msg, msg_stats_dict)
    gc.collect()

    msg_stats_dict = reset_global_msg_stats_dict(manager)
    test_queue(results, counts, msg, msg_stats_dict)
    # test_joinable_queue(results, counts, msg)
    gc.collect()
    test_pipe(results, counts, msg, msg_stats_dict)
    gc.collect()
    print_aggregated(results, counts, msg, msg_stats_dict)
    gc.collect()

"""
dell bleo
Sending 10000 numbers to Pipe() took 0.14172911643981934 seconds
Sending 100000 numbers to Pipe() took 0.5926735401153564 seconds
Sending 1000000 numbers to Pipe() took 5.1644651889801025 seconds
Sending 10000000 numbers to Pipe() took 51.538297176361084 seconds
Sending 10000 numbers to Queue() took 0.14281940460205078 seconds
Sending 100000 numbers to Queue() took 0.6834473609924316 seconds
Sending 1000000 numbers to Queue() took 6.669465065002441 seconds
Sending 10000000 numbers to Queue() took 67.16855216026306 seconds
Time delta for 10000 messages: -0.0010902881622314453[Ratio: -0.007692760595839887]
Time delta for 100000 messages: -0.0907738208770752[Ratio: -0.1531599012491889]
Time delta for 1000000 messages: -1.5049998760223389[Ratio: -0.29141446809123556]
Time delta for 10000000 messages: -15.630254983901978[Ratio: -0.3032745713428627]
Sending 10000 numbers to Queue() took 0.14954090118408203 seconds
Sending 100000 numbers to Queue() took 0.6878499984741211 seconds
Sending 1000000 numbers to Queue() took 6.7689526081085205 seconds
Sending 10000000 numbers to Queue() took 66.83499526977539 seconds
Sending 10000 numbers to Pipe() took 0.13245892524719238 seconds
Sending 100000 numbers to Pipe() took 0.6083834171295166 seconds
Sending 1000000 numbers to Pipe() took 5.1407341957092285 seconds
Sending 10000000 numbers to Pipe() took 50.90841889381409 seconds
Time delta for 10000 messages: -0.01708197593688965[Ratio: -0.12896055063870995]
Time delta for 100000 messages: -0.07946658134460449[Ratio: -0.13061924291024377]
Time delta for 1000000 messages: -1.628218412399292[Ratio: -0.31672876877359324]
Time delta for 10000000 messages: -15.926576375961304[Ratio: -0.31284759420993435]

gts_test1
Sending 10000 numbers to Pipe() took 0.02251148223876953 seconds
Sending 100000 numbers to Pipe() took 0.2347714900970459 seconds
Sending 1000000 numbers to Pipe() took 2.3396248817443848 seconds
Sending 10000000 numbers to Pipe() took 22.754943132400513 seconds
Sending 10000 numbers to Queue() took 0.03014349937438965 seconds
Sending 100000 numbers to Queue() took 0.28561949729919434 seconds
Sending 1000000 numbers to Queue() took 2.851684093475342 seconds
Sending 10000000 numbers to Queue() took 27.775741577148438 seconds
Time delta for 10000 messages: -0.007632017135620117[Ratio: -0.33902774835839866]
Time delta for 100000 messages: -0.05084800720214844[Ratio: -0.21658510231003664]
Time delta for 1000000 messages: -0.512059211730957[Ratio: -0.21886380835083885]
Time delta for 10000000 messages: -5.020798444747925[Ratio: -0.22064649494108668]
Sending 10000 numbers to Queue() took 0.03803420066833496 seconds
Sending 100000 numbers to Queue() took 0.2779529094696045 seconds
Sending 1000000 numbers to Queue() took 2.7895572185516357 seconds
Sending 10000000 numbers to Queue() took 28.34367036819458 seconds
Sending 10000 numbers to Pipe() took 0.02653670310974121 seconds
Sending 100000 numbers to Pipe() took 0.24361300468444824 seconds
Sending 1000000 numbers to Pipe() took 2.384361505508423 seconds
Sending 10000000 numbers to Pipe() took 23.647746324539185 seconds
Time delta for 10000 messages: -0.01149749755859375[Ratio: -0.433267746601619]
Time delta for 100000 messages: -0.03433990478515625[Ratio: -0.14096088519427238]
Time delta for 1000000 messages: -0.4051957130432129[Ratio: -0.1699388755048753]
Time delta for 10000000 messages: -4.6959240436553955[Ratio: -0.19857807924734255]

Zed
Sending 10000 numbers to Pipe() took 0.04478859901428223 seconds
Sending 100000 numbers to Pipe() took 0.31635212898254395 seconds
Sending 1000000 numbers to Pipe() took 3.0773861408233643 seconds
Sending 10000000 numbers to Pipe() took 30.645535469055176 seconds
Sending 10000 numbers to Queue() took 0.05045151710510254 seconds
Sending 100000 numbers to Queue() took 0.4380209445953369 seconds
Sending 1000000 numbers to Queue() took 4.3887107372283936 seconds
Sending 10000000 numbers to Queue() took 42.303247928619385 seconds
Time delta for 10000 messages: -0.0056629180908203125[Ratio: -0.12643659805064492]
Time delta for 100000 messages: -0.12166881561279297[Ratio: -0.3845993260867435]
Time delta for 1000000 messages: -1.3113245964050293[Ratio: -0.4261163651221814]
Time delta for 10000000 messages: -11.657712459564209[Ratio: -0.3804049197096188]
Sending 10000 numbers to Queue() took 0.05424952507019043 seconds
Sending 100000 numbers to Queue() took 0.44359731674194336 seconds
Sending 1000000 numbers to Queue() took 4.302424192428589 seconds
Sending 10000000 numbers to Queue() took 42.48211669921875 seconds
Sending 10000 numbers to Pipe() took 0.04413247108459473 seconds
Sending 100000 numbers to Pipe() took 0.3135397434234619 seconds
Sending 1000000 numbers to Pipe() took 3.0943334102630615 seconds
Sending 10000000 numbers to Pipe() took 31.293335914611816 seconds
Time delta for 10000 messages: -0.010117053985595703[Ratio: -0.22924286215931497]
Time delta for 100000 messages: -0.13005757331848145[Ratio: -0.41480410712343957]
Time delta for 1000000 messages: -1.2080907821655273[Ratio: -0.39042036587221635]
Time delta for 10000000 messages: -11.188780784606934[Ratio: -0.35754515961919386]

hydra
Sending 10000 numbers to Pipe() took 0.04212641716003418 seconds
Sending 100000 numbers to Pipe() took 0.3768124580383301 seconds
Sending 1000000 numbers to Pipe() took 3.6631572246551514 seconds
Sending 10000000 numbers to Pipe() took 37.24360394477844 seconds
Sending 10000 numbers to Queue() took 0.05841708183288574 seconds
Sending 100000 numbers to Queue() took 0.5189673900604248 seconds
Sending 1000000 numbers to Queue() took 5.39066481590271 seconds
Sending 10000000 numbers to Queue() took 50.12684488296509 seconds
Time delta for 10000 messages: -0.016290664672851562[Ratio: -0.38670900045842743]
Time delta for 100000 messages: -0.14215493202209473[Ratio: -0.3772564547418293]
Time delta for 1000000 messages: -1.7275075912475586[Ratio: -0.4715898022668644]
Time delta for 10000000 messages: -12.883240938186646[Ratio: -0.34591821342770124]
Sending 10000 numbers to Queue() took 0.062229156494140625 seconds
Sending 100000 numbers to Queue() took 0.5833265781402588 seconds
Sending 1000000 numbers to Queue() took 4.9200403690338135 seconds
Sending 10000000 numbers to Queue() took 51.88802647590637 seconds
Sending 10000 numbers to Pipe() took 0.053404808044433594 seconds
Sending 100000 numbers to Pipe() took 0.38507080078125 seconds
Sending 1000000 numbers to Pipe() took 3.811452627182007 seconds
Sending 10000000 numbers to Pipe() took 37.887731075286865 seconds
Time delta for 10000 messages: -0.008824348449707031[Ratio: -0.1652350934838122]
Time delta for 100000 messages: -0.1982557773590088[Ratio: -0.5148553901172928]
Time delta for 1000000 messages: -1.1085877418518066[Ratio: -0.29085701707158296]
Time delta for 10000000 messages: -14.000295400619507[Ratio: -0.3695205546301905]

stefan-box
Sending 10000 numbers to Pipe() took 0.14400625228881836 seconds
Sending 100000 numbers to Pipe() took 0.7821686267852783 seconds
Sending 1000000 numbers to Pipe() took 8.345976114273071 seconds
Sending 10000000 numbers to Pipe() took 101.45877861976624 seconds
Sending 10000 numbers to Queue() took 0.14400005340576172 seconds
Sending 100000 numbers to Queue() took 0.820995569229126 seconds
Sending 1000000 numbers to Queue() took 10.919576644897461 seconds
Sending 10000000 numbers to Queue() took 124.144357919693 seconds
Time delta for 10000 messages: 6.198883056640625e-06[Ratio: 4.3045930007317805e-05]
Time delta for 100000 messages: -0.038826942443847656[Ratio: -0.04964011737907057]
Time delta for 1000000 messages: -2.5736005306243896[Ratio: -0.3083642338998652]
Time delta for 10000000 messages: -22.685579299926758[Ratio: -0.2235940507912555]
Sending 10000 numbers to Queue() took 0.14600014686584473 seconds
Sending 100000 numbers to Queue() took 0.8171794414520264 seconds
Sending 1000000 numbers to Queue() took 15.416253805160522 seconds
Sending 10000000 numbers to Queue() took 148.11731004714966 seconds
Sending 10000 numbers to Pipe() took 0.14507722854614258 seconds
Sending 100000 numbers to Pipe() took 0.7600007057189941 seconds
Sending 1000000 numbers to Pipe() took 7.28980278968811 seconds
Sending 10000000 numbers to Pipe() took 74.57275748252869 seconds
Time delta for 10000 messages: -0.0009229183197021484[Ratio: -0.006361565691259462]
Time delta for 100000 messages: -0.05717873573303223[Ratio: -0.07523510873445653]
Time delta for 1000000 messages: -8.126451015472412[Ratio: -1.114769665232617]
Time delta for 10000000 messages: -73.54455256462097[Ratio: -0.9862120571557433]

"""


import os
import base64
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

def generate_ecc_keys():
  """
  Generates an ECC public/private key pair.

  Returns:
    tuple: A tuple containing the private and public keys.
  """
  private_key = ec.generate_private_key(curve=ec.SECP256K1())
  public_key = private_key.public_key()
  return private_key, public_key



def __derive_shared_key(private_key, peer_public_key):
  """
  Derives a shared key using own private key and peer's public key.

  Parameters
  ----------
  private_key : cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey
      The private key to use for derivation.
  peer_public_key : cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey
      The peer's public key.
  
  Returns
  -------
  bytes
      The derived shared key.
  """
  shared_key = private_key.exchange(ec.ECDH(), peer_public_key)
  derived_key = HKDF(
    algorithm=hashes.SHA256(),
    length=32,
    salt=None,
    info=b'handshake data',
    backend=default_backend()
  ).derive(shared_key)
  return derived_key

def encrypt(sender_sk, receiver_pk, plaintext):
  """
  Encrypts plaintext using the sender's private key and receiver's public key, 
  then base64 encodes the output.

  Parameters
  ----------
  sender_sk : cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey
      The sender's private key for deriving the shared key.
      
  receiver_pk : cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey
      The receiver's public key for deriving the shared key.
      
  plaintext : str
      The plaintext to encrypt.

  Returns
  -------
  str
      The base64 encoded nonce and ciphertext.
  """
  shared_key = __derive_shared_key(sender_sk, receiver_pk)
  aesgcm = AESGCM(shared_key)
  nonce = os.urandom(12)  # Generate a unique nonce for each encryption
  ciphertext = aesgcm.encrypt(nonce, plaintext.encode(), None)
  encrypted_data = nonce + ciphertext  # Prepend the nonce to the ciphertext
  print(f"Encrypted Data: {encrypted_data}")
  print(f"Nounce: {nonce}")
  return base64.b64encode(encrypted_data).decode()  # Encode to base64

def decrypt(receiver_sk, sender_pk, encrypted_data_b64):
  """
  Decrypts base64 encoded encrypted data using the receiver's private key.

  Parameters
  ----------
  receiver_sk : cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey
      The receiver's private key for deriving the shared key.
      
  sender_pk : cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey
      The sender's public key for deriving the shared key.
      
  encrypted_data_b64 : str
      The base64 encoded nonce and ciphertext.

  Returns
  -------
  str
      The decrypted plaintext.

  """
  encrypted_data = base64.b64decode(encrypted_data_b64)  # Decode from base64
  nonce = encrypted_data[:12]  # Extract the nonce
  ciphertext = encrypted_data[12:]  # The rest is the ciphertext
  print(f"Encrypted Data: {encrypted_data}")
  print(f"Nonce: {nonce}")
  shared_key = __derive_shared_key(receiver_sk, sender_pk)
  aesgcm = AESGCM(shared_key)
  plaintext = aesgcm.decrypt(nonce, ciphertext, None)
  return plaintext.decode()

# Example usage comments would indicate how to use these functions with base64 encoded data.

if __name__ == "__main__":
  import os

  # Generate key pairs for two parties (simulating sender and receiver)
  sender_private_key, sender_public_key = generate_ecc_keys()
  receiver_private_key, receiver_public_key = generate_ecc_keys()
  # bandit is an eavesdropper
  bandit_private_key, bandit_public_key = generate_ecc_keys()

  message = "Hello, receiver!"
  encrypted_message = encrypt(
    sender_sk=sender_private_key, 
    receiver_pk=receiver_public_key, 
    plaintext=message,
  )
  print("Encrypted Message:  {}".format(encrypted_message))

  decrypted_message = decrypt(
    receiver_sk=receiver_private_key, 
    sender_pk=sender_public_key,
    encrypted_data_b64=encrypted_message,
  )

  try:
    # Bandit tries to decrypt the message
    bandit_decrypted_message = decrypt(
      receiver_sk=bandit_private_key, 
      encrypted_data_b64=encrypted_message,
    )
  except:
    bandit_decrypted_message = "Decryption failed"

  print(f"Encrypted Message:  {encrypted_message}")
  print(f"Decrypted Message:  {decrypted_message}")
  print(f"Bandit Message:     {bandit_decrypted_message}")
